import numpy as np
import os
import nibabel as nib
import vnmrjpy as vj
from shutil import copyfile

class SaveRecon():
    """Save reconstruction results to nifti or fdf files in orderly manner

    Args:
        kspace
        imagespace
        procpar

    Methods:

        save
    """
    def __init__(self, procpar, kspace=None, imagespace=None):

        self.imagespace = imagespace
        self.kspace = kspace
        self.ppdict = vj.io.ProcparReader(procpar).read()
        self.procpar = procpar
        self.basename = self.ppdict['pslabel']+'_'+\
                        self.ppdict['comment']

        #raise(Exception('Please specifiy either kspace data or\
        #                    image data, or both'))

    def save(self, outdir, savetype='full',filetype='nifti'):
        """Saves desired files to output dir

        Creates directory, if not present

        Args:
            outdir -- path/to/saved/files
        Options:
            savetype -- determines what to save

                'full' -- combined mangitude,per channel mangitude, phase,
                          per channel kspace real, imag part
                'magn' -- only combined magnitude
                'fullimag'  -- combined magnitude, per channel magnitude,phase
                'fullkspace' -- per channel imag, real kspace
                'mask' -- only save kspace mask

            filetype -- output file type

                'nifti'
                'fdf'
        """
        if not os.path.exists(outdir):
            try:
                os.makedirs(outdir)
            except:
                raise

        copyfile(self.procpar,outdir+'/procpar')

        if savetype in ['magn','fullimag','full']:
            #saving sum of squares magnitude
            img = np.sum(np.absolute(self.imagespace),axis=0)
            out = outdir+'/'+self.basename+'_magn'
            vj.io.NiftiWriter(img,self.procpar).write(out)
        if savetype in ['fullimag','full']:
            # saving magnitude for each channel
            for ch in range(self.imagespace.shape[0]):
                img = np.absolute(self.imagespace[ch,...])
                out = outdir+'/'+self.basename+'_magn_ch'+str(ch+1)
                vj.io.NiftiWriter(img,self.procpar).write(out)
            # saving phase for each channel
            for ch in range(self.imagespace.shape[0]):
                img = np.arctan2(np.imag(self.imagespace[ch,...]),\
                                np.real(self.imagespace[ch,...]))
                out = outdir+'/'+self.basename+'_phase_ch'+str(ch+1)
                vj.io.NiftiWriter(img,self.procpar).write(out)
        if savetype in ['fullkspace','full']:
            # saving real part of kspace
            for ch in range(self.kspace.shape[0]):
                img = np.real(self.kspace[ch,...])
                out = outdir+'/'+self.basename+'_kspace_ch'+str(ch+1)+'_real'
                vj.io.NiftiWriter(img,self.procpar).write(out)
            # saving imag part of kspace
            for ch in range(self.kspace.shape[0]):
                img = np.imag(self.kspace[ch,...])
                out = outdir+'/'+self.basename+'_kspace_ch'+str(ch+1)+'_imag'
                vj.io.NiftiWriter(img,self.procpar).write(out)
        if savetype == 'full':
            pass
        if savetype == 'mask':
            # saving only mask
            img = self.imagespace
            out = outdir+'/'+self.basename+'_mask'
            vj.io.NiftiWriter(img,self.procpar).write(out)







import vnmrjpy as vj
import unittest
import glob

class Test_Composer(unittest.TestCase):

    def test_composer(self):

        compdir = vj.config['dataset_dir']+'/parameterfit/composer'
        ref = glob.glob(compdir+'/*composer08_01.fid')[0]
        mems = glob.glob(compdir+'/mems_s_2019022301_01*')[0]
        procpar_ref = ref+'/procpar'
        procpar = mems+'/procpar'
        #make images
        (data_ref, hdr_ref) = vj.io.FidReader(ref).read()
        (data, hdr) = vj.io.FidReader(mems).read()
        kspace_ref = vj.recon.KspaceMaker(data_ref, hdr_ref, procpar_ref).make()
        kspace = vj.recon.KspaceMaker(data, hdr, procpar).make()
        imgspace = vj.recon.ImageSpaceMaker(kspace, procpar).make()
        imgspace_ref = vj.recon.ImageSpaceMaker(kspace_ref, procpar_ref).make()

        #imgspace = vj.util.to_scanner_space(imgspace,procpar)
        #imgspace_ref = vj.util.to_scanner_space(imgspace_ref,procpar_ref)
        
        comp = vj.recon.Composer(imgspace, imgspace_ref, procpar, procpar_ref)
        comp.flirt_registration()
        



import vnmrjpy as vj
import numpy as np
import unittest
import glob
import matplotlib.pyplot as plt

class Test_core_to_imagespace(unittest.TestCase):

    def test_gems(self):

        seq = glob.glob(vj.config['fids_dir']+'/gems*axial_*')[0]
        varr = vj.core.read_fid(seq)
        varr.to_kspace().to_imagespace()
        #plt.imshow(np.absolute(varr.data[:,:,10,0,1]))
        #plt.show()
        self.assertEqual(varr.vdtype,'imagespace')

    def test_ge3d(self):

        seq = glob.glob(vj.config['fids_dir']+'/ge3d*axial_*')[0]
        varr = vj.core.read_fid(seq)
        varr.to_kspace().to_imagespace()
        self.assertEqual(varr.vdtype,'imagespace')



import vnmrjpy as vj
import os
import json

"""
Collection of unsorted utility functions. Includes:
    
    vprint -- print if verbose is True
    savepd -- save procpar dictionary to json
    loadpd -- load procpar dictionary from json

"""

def vprint(string):

    if vj.config['verbose']==True:
        print(string)
    else:
        pass


def getpetab(pd):

    pass 

def savepd(pd,out):
    """Save procpar dictionary to json file"""
    # enforce .json 
    if out.endswith('.json'):
        pass
    else:
        out = out+'.json'
    # check basedir:
    basedir = out.rsplit('/',1)[0]
    if not os.path.exists(basedir):
        os.makedirs(basedir)
    with open(out,'w') as openfile:
        jsondata = json.dumps(pd)
        json.dump(pd,openfile)

def loadpd(infile):
    """Load procpar dictionary from json file"""
    with open(infile, 'r') as openfile:
        pd = json.load(openfile)
    return pd



"""
read
====

Read data from various formats, including Varian specific ones:
    
Varian: fid, fdf, procpar
General Nifti1: nii, nii.gz

"""
import vnmrjpy as vj
from vnmrjpy.core.utils import vprint
import numpy as np
import os
import glob
import csv

def read_procpar(procpar):
    """Return dictionary of varian parameters from procpar file"""

    with open(procpar,'r') as openpp:
        
        name = []
        value = []
        subtype = []
        basictype = []
        maxvalue = []
        minvalue = []
        # possibilities (as of manual): 0 :'line1', 1: 'line2', 2: 'last'
        line_id = 0 
    
        current_name = 0

        for line_num, line in enumerate(openpp.readlines()):

            if line_id == 0:  # parameter name line
                
                fields = line.rsplit('\n')[0].rsplit(' ')
                current_name = fields[0]
                name.append(fields[0])
                subtype.append(int(fields[1]))
                basictype.append(int(fields[2]))

                # can be 0 (undefined), 1 (real), 2 (string)
                current_basictype = int(fields[2])  
                val_list = []  # if parameter value is a multitude of strings
                line_id = 1
                                
            elif line_id == 1:  # parameter value line

                fields = line.rsplit('\n')[0].rsplit(' ')

                # values are real, and all are on this line
                if current_basictype == 1:  
                    val = [fields[i] for i in range(1,len(fields)-1)]
                    if len(val) == 1:  # don't make list if there is only one value
                        val = val[0]
                    value.append(val)
                    line_id = 2
                # values are strings
                elif current_basictype == 2 and int(fields[0]) == 1:
                    
                    value.append(str(fields[1].rsplit('"')[1]))
                    line_id = 2
                # multiple string values
                elif current_basictype == 2 and int(fields[0]) > 1:
                    
                    val_list.append(fields[1])
                    remaining_values = int(fields[0])-1
                    line_id = 3
                
            elif line_id == 2:

                line_id = 0
        
            elif line_id == 3:  # if values are on multiple lines
                if remaining_values > 1:  # count backwards from remaining values
                    val_list.append(fields[0])
                    remaining_values = remaining_values - 1
                elif remaining_values == 1:
                    val_list.append(fields[0])
                    remaining_values = remaining_values - 1
                    value.append(val_list)
                    line_id = 2
                else:
                    raise(Exception('Error reading procpar, problem with line_id'))
            else:
                raise(Exception('Error reading procpar, problem with line_id'))

    vprint('procpar file {} read succesfully'.format(procpar))

    return dict(zip(name, value))

def read_fid(fid,procpar=None):
    """Handles raw data from Varian spectrometer

    .fid File structure as per Vnmrj manual:
    ===================================
    struct datafilehead
    Used at the beginning of each data file (fid's, spectra, 2D) 

       int     nblocks;      /* number of blocks in file
       int     ntraces;      /* number of traces per block    
       int     np;           /* number of elements per trace    
       int     ebytes;       /* number of bytes per element            
       int     tbytes;       /* number of bytes per trace        
       int     bbytes;       /* number of bytes per block        
       short   vers_id;      /* software version and file_id status bits
       short   status;       /* status of whole file    
       int       nbheaders;     /* number of block headers            

    struct datablockhead
     Each file block contains the following header 

       short   scale;    /* scaling factor        
       short   status;    /* status of data in block      
       short   index;    /* block index              
       short   mode;    /* mode of data in block     
       int       ctcount;    /* ct value for FID        
       float   lpval;    /* F2 left phase in phasefile    
       float   rpval;    /* F2 right phase in phasefile
       float   lvl;        /* F2 level drift correction       
       float   tlt;        /* F2 tilt drift correction      
    """

    def _decode_header(bheader):
        """Return dictionary of fid header from binary header data"""
        hkey_list = ['nblocks','ntraces','np','ebytes','tbytes','bbytes',\
                 'vers_id','status','nbheaders']
        hval_list = []
        h_dict = {}
        if len(bheader) != 32:
            raise(Exception('Incorrect fid header data: not 32 bytes'))
        else: 
            hval_list.append(int.from_bytes(bheader[0:4],byteorder='big'))
            hval_list.append(int.from_bytes(bheader[4:8],byteorder='big'))
            hval_list.append(int.from_bytes(bheader[8:12],byteorder='big'))
            hval_list.append(int.from_bytes(bheader[12:16],byteorder='big'))
            hval_list.append(int.from_bytes(bheader[16:20],byteorder='big'))
            hval_list.append(int.from_bytes(bheader[20:24],byteorder='big'))
            hval_list.append(int.from_bytes(bheader[24:26],byteorder='big'))
            hval_list.append(int.from_bytes(bheader[26:28],byteorder='big'))
            hval_list.append(int.from_bytes(bheader[28:],byteorder='big'))

            for num, i in enumerate(hkey_list):
                h_dict[i] = hval_list[num]
        
        return h_dict

    def _decode_blockhead(blockheader):
        """Return block header dictionary"""
        bh_dict = {}
        bhk_list = ['scale','status','index','mode','ctcount','lpval'\
                    'rpval','lvl','tlt']
        bhv_list = []
        bhv_list.append(int.from_bytes(blockheader[0:2],byteorder='big'))
        bhv_list.append(int.from_bytes(blockheader[2:4],byteorder='big'))
        bhv_list.append(int.from_bytes(blockheader[4:6],byteorder='big'))
        bhv_list.append(int.from_bytes(blockheader[6:8],byteorder='big'))
        bhv_list.append(int.from_bytes(blockheader[8:12],byteorder='big'))
        bhv_list.append(int.from_bytes(blockheader[12:16],byteorder='big'))
        bhv_list.append(int.from_bytes(blockheader[16:20],byteorder='big'))
        bhv_list.append(int.from_bytes(blockheader[20:24],byteorder='big'))
        bhv_list.append(int.from_bytes(blockheader[24:28],byteorder='big'))

        for num, i in enumerate(bhk_list):
            bh_dict[i] = bhv_list[num]        
        return bh_dict

    def _chunks(l, n):
        """Generate n long chunks from list l"""
        for i in range(0, len(l), n):
            yield l[i:i + n]

    # init
    if procpar==None:
    
        if os.path.isdir(fid):
            procpar = str(fid)+'/procpar'
            fid = str(fid)+'/fid'
        else:
            procpar = fid.rsplit('/')[0]+'/procpar'

    # header data should be 32 bytes
    with open(fid,'rb') as openFid:
        binary_data = openFid.read()
        # binary header data 
        bheader = bytearray(binary_data[:32])
        # binary fid data
        bdata = binary_data[32:]

    header_dict = _decode_header(bheader)
    pd = read_procpar(procpar)

    chunk_size = int(header_dict['bbytes'])
    block_list = list(_chunks(bdata,chunk_size))

    if header_dict['bbytes'] == len(block_list[0]):
        #print('blocksize check OK')
        pass

    vprint('reading fid {}'.format(fid))
    vprint('\nfid header :\n')
    vprint(header_dict)

    #------------main iteration through bytearray -----------------

    dim = (int(header_dict['nblocks']),\
            int((header_dict['ntraces'])*int(header_dict['np'])))
    fid_data = np.empty(dim)
    if header_dict['ebytes'] == 4:
        dt = '>f'
    if header_dict['ebytes'] == 2:
        dt = '>i2'
    for k,block in enumerate(block_list): # for each block
        block_header = block[:28] # separate block header
        block_data = block[28:]
        fid_data[k,:] = np.frombuffer(bytearray(block_data),dt)

    # fid data ready, now create varray class
    arr = _get_arrayed_par_length(pd)
    sdim = ['phase', 'read', 'slice', 'time', 'rcvr']
    return  vj.varray(data=fid_data,space=None,pd=pd,fid_header=header_dict,\
                        source='fid',dtype=vj.DTYPE, seqcon=pd['seqcon'],\
                        apptype=pd['apptype'],arrayed_params=arr,vdtype='fid',\
                        sdims = sdim)

def read_fdf(path):
    """Return vnmrjpy.varray from varian .fdf files

    Input can be the whole .img directory (this is preferred) or stand-alone
    .fdf files. Procpar file should be present in directory.

    Args:
        path
    Return:
        varray
    """
    #---------------- auxiliary functions for read method -----------------

    def _preproc_fdf(fdf):

        with open(fdf,'rb') as openFdf:
            fdata = bytearray(openFdf.read())
            nul = fdata.find(b'\x00')
            header = fdata[:nul]
            data = fdata[nul+1:]
        return (header,data)

    # ----------parse fdf header and return into a dictionary --------------

    def _parse_header(header):
        keys_to_parse = sorted(['rank','roi','location','spatial_rank',\
                        'matrix','orientation',\
                        'studyid','gap','pe_size','ro_size',\
                        'pe2_size', 'abscissa',\
                        'storage'])
        to_delete = ('char','float','int')
        header = header.decode('ascii').split('\n')
        header_dict = {}    
        for line in header:             # some formatting of header
            vprint(line)
            for item in to_delete:
                if line.startswith(item):
                    line = line.split(item,1)
                    break
            try:
                line = line[1].lstrip()
                line = line.lstrip('*').rstrip(';')
                if '[]' in line:
                    line = line.replace('[]','')
                if '{' in line:
                    line = line.replace('{','(')
                if '}' in line:
                    line = line.replace('}',')')
                if ' ' in line:
                    line = line.replace(' ','')
                line = line.split('=')
                header_dict[line[0]] = line[1]        
            except:
                continue

        for item in keys_to_parse:
            if item in header_dict.keys():            
                if item == 'abscissa':
                    tempval = header_dict[item][1:-1];''.join(tempval)
                    tempval = tempval.replace('"','')                
                    header_dict[item] = tuple([k \
                                    for k in tempval.split(',')])
                if item == 'matrix':            
                    tempval = header_dict[item][1:-1];''.join(tempval)                
                    header_dict[item] = tuple([int(k) \
                                    for k in tempval.split(',')])
                if item == 'roi':
                    tempval = header_dict[item][1:-1];''.join(tempval)                
                    header_dict[item] = tuple([float(k)\
                                    for k in tempval.split(',')])
                if item == 'ro_size' or item == 'pe_size' \
                                    or item == 'pe2_size':
                    header_dict[item] = int(header_dict[item])
                if item == 'storage':
                    tempval = header_dict[item];''.join(tempval)
                    tempval = tempval.replace('"','')
                    header_dict[item] = str(tempval)
                if item == 'orientation':
                    tempval = header_dict[item][1:-1];''.join(tempval)                
                    header_dict[item] = tuple([float(k)\
                                    for k in tempval.split(',')])
                if item == 'location':
                    tempval = header_dict[item][1:-1];''.join(tempval)                
                    header_dict[item] = tuple([float(k)\
                                    for k in tempval.split(',')])
                if item == 'gap':
                    header_dict[item] = float(header_dict[item])
                if item == 'slices':
                    header_dict[item] = int(header_dict[item])
                if item == 'TR':
                    header_dict[item] = float(header_dict[item])/1000
        return header_dict

    #----------process bynary data based on header--------------------

    def _prepare_data(binary_data, header_dict):

        matrix = header_dict['matrix']

        if header_dict['storage'] == 'float' and \
           header_dict['bits'] == '32':
            dt = np.dtype('float32'); dt = dt.newbyteorder('<')

        else:
            raise(Exception('bits may be incorrectly specified in header data'))

        img_data = np.frombuffer(binary_data, dtype=dt)
        img_data = np.reshape(img_data,matrix)
        return img_data
    #---------------------------------------------------------------------
    #                             main read method
    #----------------------------------------------------------------------
    if os.path.isdir(path):
        vprint('\nMaking varray from fdf :  {} \n'.format(path))
        procpar = str(path)+'/procpar'
        fdf_list = sorted(glob.glob(str(path)+'/*.fdf'))
        (header, data) = _preproc_fdf(fdf_list[0])
        pd = read_procpar(procpar)
    else:
        if not path.endswith('.fdf'):
            warnings.warn('Input does not end with .fdf, path might be incorrect')
        vprint('\nMaking varray from fdf :  {} \n'.format(path))
        procpar = path.rsplit('/')[0]+'/procpar'
        pd = read_procpar(procpar)
        fdf_list =[path]
        (header, data) = _preproc_fdf(path)
    
    header_dict = _parse_header(header)
    # ------------------------process if 3d --------------------
    if header_dict['spatial_rank'] == '"3dfov"':

        full_data = []
        time_concat = []
        time = len([1 for i in fdf_list if 'slab001' in i])
        for i in fdf_list: # only 1 item, but there migh be more in future
                            
            (header, data) = _preproc_fdf(i)
            img_data = _prepare_data(data, header_dict)
            full_data.append(img_data) # full data in one list

        data_array = np.asarray(full_data)

    #----------------------process if 2d-------------------------
    elif header_dict['spatial_rank'] == '"2dfov"':

        full_data = []
        time_concat = []
        time = len([1 for i in fdf_list if 'slice001' in i])

        for i in fdf_list:                
            (header, data) = _preproc_fdf(i)
            img_data = _prepare_data(data,header_dict)
            # expand 2d to 4d
            img_data = np.expand_dims(np.expand_dims(img_data,2),3)
            full_data.append(img_data) # full data in one list
        # make sublists
        slice_list =[full_data[i:i+time] \
                        for i in range(0,len(full_data),time)]
        for aslice in slice_list:
            time_concat.append(np.concatenate(tuple(aslice),axis=3))
        #slice+time concatenated
        slice_time_concat = np.concatenate(time_concat, axis=2)
        data_array = slice_time_concat
    # --------------------process if 1d------------------------
    elif header_dict['spatial_rank'] == '"1dfov"':
        raise(Exception('Not implemented'))
     
    # making vnmrjpy.varray
    arrayed_params = _get_arrayed_par_length(pd) 
    
    varr = vj.varray(data=data_array, fdf_header=header_dict, pd=pd,\
                    dtype=vj.DTYPE, source='fdf', seqcon=pd['seqcon'],\
                    apptype=pd['apptype'],arrayed_params=arrayed_params,\
                    vdtype='image',space=None)
    
    varr.space = None
    varr.set_nifti_header()
    varr.to_local()
    return varr

def _get_arrayed_par_length(pd):
    """Return tuple of (name, length) of arrayed acquisition parameters

    List of possible arrayed parameters should be updated in config
    """
    pars = vj.config['array_pars']
    lengths = []
    for par in pars:
        if type(pd[par]) == list:
            l = len(pd[par])
        else:
            l = 1
        lengths.append(l)
    arrayed_pars = [(pars[i],lengths[i]) for i in range(len(pars))]
    return arrayed_pars

def read_petab(ppdict):
    """Get and parse phase encode table for sequence

    Args:
        ppdict -- procpar dictionary

    Return:
        petab (np.ndarray)
    """
    def _readpetabfile(petabfile):
        """Parses petab file contents into appropriate array"""
        lst = []
        lst_fin = []
        if os.path.exists(petabfile):
            pass
        else:
            raise(Exception('Could not find petab file'))
        with open(petabfile, 'r') as openfile:
            reader = csv.reader(openfile,delimiter='\t')
            for item in reader:
                lst.append(item)
            # removing 't1 ='
            lst = lst[1:]
            for item in lst:
                item_fin = []
                for element in item[:-1]:  # the last one is whitespace
                    element = int(element.replace(" ",""))
                    item_fin.append(element)
                lst_fin.append(item_fin)

        petab = np.array(lst_fin)
        return petab

    try:
        petabfile_name = ppdict['petable']
    except:
        return None
    try:
        fullpath = vj.config['tablib_dir']+'/'+petabfile_name
        # TODO check file name consistency
        return _readpetabfile(fullpath)
    except:
        raise(Exception('could not find petab file'))



import os
import glob
import nibabel as nib
import numpy as np

class Ideal():
    """Fat fraction calculation for Dixon method with iterative least squares fitting.

    Ref.: dixon paper, ideal paper  
    """
    def __init__(self,kspace_data_list, ro_shift_list, freq=(0,1400)):
        """Fat fraction calculation for Dixon method with iterative least squares fitting.

        Args:
            kspace_data_list (list of np.ndarra) -- k-spaces with different
                                                    readout shifts
            ro_shift_list (tuple) -- readout shifts in seconds
            freq (tuple) -- frequency offsets in Hz of the Dixon components at 9.4T
        """

        self.data_list = kspace_data_list
        self.roshift_lsit = ro_shift_list
        self.freq = freq

    def fit(self, iterations, mask=None, fieldcorr=None, selfmask=False):
        """IDEAL fitting method
        
        Args.:
            iterations
            mask (np.ndarray) -- binary mask where fitting is performed

        Return:
        """

        def apply_field_correction(data, field, roshift):

            for k, shift in enumerate(roshift):
                data[k,...] = data[k,...]*np.exp(-1j*2*np.pi*field*shift)            
            return data

        def recalc_ro_final(ro, A, field, S):
        
            fullA = np.repeat(A,S.size).reshape(A.shape + S.shape,order='c')
            ro =  np.linalg.multi_dot((np.linalg.inv(np.dot(A.T,A)),A.T,S))
            return ro

        def make_Acd(roshift, freq):
            A = np.zeros((2*len(roshift),2*len(freq)))
            c = np.asarray([math.cos(2*np.pi*i*j) for i in freq for j in roshift])
            d = np.asarray([math.sin(2*np.pi*i*j) for i in freq for j in roshift])
            c = np.reshape(c,(len(freq),len(roshift)),order='c')
            d = np.reshape(d,(len(freq),len(roshift)),order='c')
            
            A[:len(roshift),0::2] = c.T
            A[len(roshift):,0::2] = d.T
            A[:len(roshift),1::2] = -d.T
            A[len(roshift):,1::2] = c.T 

            return A, c, d

        def fit_voxelvise(data1D, roshift, A, c, d, F0=0, freq=self.freq):
            """1d function to be used for apply_along_axis
    
            F0 is initial field
            freq is the Dixon components frequency shift
            """
            def _Sh_from_f(data1D, field, freq, roshift):

                Sh = data1D*np.exp(-1j*2*np.pi*field*np.asarray(roshift))
                Sre = np.asarray([np.real(Sh[i]) for i in range(Sh.shape[0])])
                Sim = np.asarray([np.imag(Sh[i]) for i in range(Sh.shape[0])])

                return np.concatenate((Sre, Sim),axis=0).T # return col vector
            
            def _ro_from_Sh(S,A):

                return np.linalg.multi_dot((np.linalg.inv(np.dot(A.T,A)),A.T,S))

            def _B_from_ro(ro, A, c, d, roshift, freq):

                B = np.zeros((A.shape[0],A.shape[1]+1))
                B[:,1:] = A
                rore = ro[0::2]
                roim = ro[1::2]
                gre = [2*np.pi*roshift[i]*(-rore[j]*d[j,i]-roim[j]*c[j,i]) \
                         for i in range(len(roshift)) for j in range(len(freq))] # g1N_re
                gre = np.sum(np.reshape(gre,(c.shape),order='F'),axis=0)
                #gre = np.reshape(gre,(c.shape),order='F')
                gim = [2*np.pi*roshift[i]*(rore[j]*c[j,i]-roim[j]*d[j,i]) \
                         for i in range(len(roshift)) for j in range(len(freq))] # g1N_re
                gim = np.sum(np.reshape(gim,(c.shape),order='F'),axis=0)
                #gre = np.reshape(gre,(c.shape),order='F')
            
                B[:len(gre),0] = gre
                B[len(gim):,0] = gim

                return B

            def _y_from_Sh(S,B):# y contains the error terms delta_field, delta_ro     

                return np.linalg.multi_dot((np.linalg.inv(np.dot(B.T,B)),B.T,S))

            def _Sh_from_B_y(B,y):
    
                return np.dot(B,y)

            # actual iteration starts here

            f = 0 # init fieldmap=0
            for i in range(iterations):

                Sh = _Sh_from_f(data1D,f,freq,roshift)
                ro = _ro_from_Sh(Sh,A)
                B = _B_from_ro(ro,A,c,d,roshift,freq)
                y = _y_from_Sh(Sh,B)
                f = np.asarray(f + y[0]) # recalculate field
                # make iteration stop on voxel basis
                if abs(y[0]) < 1:
                    break

            return np.concatenate([ro,y,np.atleast_1d(f)])

        """-------------------------IDEAL_fit MAIN-------------------------------------"""

        if self.proc_check_ok:
            print('dixonProc.ideal_fit() : Preprocess check OK')
        else:
            print('dixonProc.ideal_fit() : Preprocess was not done. \
                    Use dixonPreproc().run() first')
            return
        # data.shape : [roshift, read, phase, slice, time=1]
        # field.shape : [read, phase, slice, 1]

        data, field, roshift, header, affine = read_proc_dir()

        if fieldcorr == True: 

            data = apply_field_correction(data, field, roshift)

        A,c,d = make_Acd(roshift, self.freq)

        out = np.apply_along_axis(fit_voxelvise, 0, data, roshift, A, c, d, self.freq)

        if self.saving:
            self.save_results_to_nifti(out, affine, header)
        print('Done!')

        return out # data is the masked data

if __name__ == '__main__':
    
    study_dir = '/home/david/dev/dixon/s_2018080901'

    dp = dixonProc(study_dir)
    #dp.least_squares_fit()
    dp.ideal_fit(100)



import vnmrjpy as vj
import numpy as np
from scipy.optimize import least_squares
from scipy.optimize import curve_fit
import itertools
import copy
import matplotlib.pyplot as plt
import sys
from lmfit import minimize, Parameters

def z_spectrum(delta_w, b1, dw, c, d, tp=0.005, f_scale=400, gamma=42.57747):
    """Return curve of Z spectrum for WASABI

    Args:
        delta_w -- magnetisation transfer offset frequencies
        b1 -- B1 transverse field in microT
        dw -- freq offset made by B0 offset
        c -- fit parameter
        d -- fit parameter
        
        tp -- offset pulse duration
        gamma -- giromagnetic ratio for H1
        f_scale -- frequency scaling factor (only for easier parameter fitting)
    """
    z = np.absolute(c - d*np.sin(np.arctan((gamma*b1)/delta_w-dw*f_scale))**2\
         *np.sin(np.sqrt((gamma*b1)**2+(delta_w-dw*f_scale)**2) * tp/2)**2)
    return z

def WASABI(varr, norm, mask='none'):
    """WASABI B0 and B1 mapping function

    Args:
        varr (vj.varray) -- input varray to fit
        norm (vj.varray) -- additional data for normalization
        mask (np.ndarray) -- data mask where fitting should be done
    Return:
        b0map (vj.varray) -- 
        b1map (vj.varray) -- 
        c (vj.varray) -- fit parameter
        d (vj.varray) -- fit parameter

    Use with specifically designed sequence. See reference paper for theory.

    Ref.: Schuenke et al.: Simultaneous mapping of water shift and B1
        (WASABI) - Application to field-inhomogenty correction of CESTMRI
        data.
    """
    # -----------------------helper fuctions-----------------------------------
    
    def _residual(params, x, data, eps_data):
        """Residual for least-squares fitting"""
        b1 = params['b1']    
        dw = params['dw']    
        c = params['c']    
        d = params['d'] 
        model = z_spectrum(x,b1,dw,c,d,tp=tp,f_scale=f_scale)
        return (data-model) / eps_data   


    # checking whether data is actually used for wasabi
    if len(varr.pd['mtfrq']) == 1:
        raise(Exception('Only one MT frequence is found. Cannot use WASABI'))
    # checking for imagespace
    if varr.vdtype != 'imagespace':
        print('Waring: Input varray is not in imagespace.'
                'Trying to_imagespace() and proceeding...')
        varr.to_imagespace()
   
    #-------------------------- actual satart----------------------------------

    gamma = 42.57747  # giromagnetic ratio in Hz/T
    tp = float(varr.pd['pmt'])/10**6  # is in microsec initially, set to sec
    flip1 = float(varr.pd['flip1'])*2*np.pi/360  #flip angle set to rad
    flipmt = float(varr.pd['flipmt'])*2*np.pi/360  #mt flip angle set to rad
    B1 = flipmt / (2*np.pi*gamma*tp) # B1 in microT
    eps_data = 1 #TODO estimated uncertainty of data. get from noise
    delta_w = []  # freq offsets
    for item in varr.pd['mtfrq']:
        delta_w.append(float(item))
    f_scale = 400  # frequency scaling, jsut for parameter estimation help

    print('flip1 {}'.format(flip1))
    print('tp : {}'.format(tp))
    print('B1 : {}'.format(B1))

    # make magnitude image
    magn_orig = vj.core.recon.ssos(varr.data) 
    #magn = copy.deepcopy(magn)
    magn_norm = vj.core.recon.ssos(norm.data)
    #mask if available
    if mask != 'none':
        magn_orig = magn_orig*mask
    # normalize data
    magn = np.divide(magn_orig,magn_norm) 
    # init output
    B0map = np.zeros_like(magn)
    B1map = np.zeros_like(magn)
    c_map = np.zeros_like(magn)
    d_map = np.zeros_like(magn)
    # std error maps
    B0map_err = np.zeros_like(magn)
    B1map_err = np.zeros_like(magn)
    c_map_err = np.zeros_like(magn)
    d_map_err = np.zeros_like(magn)
    # setting up lmfit
    params = Parameters()
    params.add('b1',value=3.8,min=0.5,max=7)
    params.add('dw',value=0,min=-2,max=2)
    params.add('c',value=0.5,min=0.1,max=1)
    params.add('d',value=1,min=0.3,max=3)
    # iterate over every voxel in space
    x_dim, y_dim, z_dim = varr.data.shape[0:3]
    for x, y, z in itertools.product(*map(range, (x_dim,y_dim,z_dim))):
        if mask[x,y,z,0] == 0:
            continue
        else:
            xdata = delta_w
            ydata = magn[x,y,z,:]
            try:
                out = minimize(_residual, params, args=(xdata,ydata,eps_data))
                # output vals
                B1map[x,y,z,0] = out.params['b1'].value
                B0map[x,y,z,0] = out.params['dw'].value
                c_map[x,y,z,0] = out.params['c'].value
                d_map[x,y,z,0] = out.params['d'].value
                # output errors
                B1map_err[x,y,z,0] = out.params['b1'].stderr
                B0map_err[x,y,z,0] = out.params['dw'].stderr
                c_map_err[x,y,z,0] = out.params['c'].stderr
                d_map_err[x,y,z,0] = out.params['d'].stderr
            except Exception:
                continue
            except KeyboardInterrupt:
                sys.exit(0)
    # creating varray
    varr_b0 = copy.copy(norm)
    varr_b1 = copy.copy(norm)
    varr_c = copy.copy(norm)
    varr_d = copy.copy(norm)
    # create error varrs
    varr_b0_err = copy.copy(norm)
    varr_b1_err = copy.copy(norm)
    varr_c_err = copy.copy(norm)
    varr_d_err = copy.copy(norm)

    varr_b0.data = np.array(B0map,dtype='float32')
    varr_b0.intent = 'B0_map'
    varr_b1.data = np.array(B1map,dtype='float32')
    varr_b1.intent = 'B1_map'
    varr_c.data = np.array(c_map,dtype='float32')
    varr_c.intent = 'wasabi_c'
    varr_d.data = np.array(d_map,dtype='float32')
    varr_d.intent = 'wasabi_d'
    # errors
    varr_b0_err.data = np.array(B0map_err,dtype='float32')
    varr_b1_err.data = np.array(B1map_err,dtype='float32')
    varr_c_err.data = np.array(c_map_err,dtype='float32')
    varr_d_err.data = np.array(d_map_err,dtype='float32')
    return ((varr_b0, varr_b1, varr_c, varr_d) , \
                (varr_b0_err, varr_b1_err, varr_c_err, varr_d_err))



import unittest
import vnmrjpy as vj
import glob
import numpy as np

vj.config['verbose']=False

class Test_Core(unittest.TestCase):

    def test_core_basic(self):

        shape = (128,128,128,5,4)
        data = np.random.rand(*shape)
        varr = vj.varray(data=data)
        self.assertEqual(varr.data.shape,shape)

    def test_core_arrayed_procpar(self):

        gems_arr_te = glob.glob(vj.config['dataset_dir']+'/arrayed/'+\
                            'gems_*array_te*')[0]
        procpar = gems_arr_te+'/procpar'
        pd = vj.read_procpar(procpar)
        varray = vj.read_fid(gems_arr_te)
        #self.assertEqual(varray.arrayed_params)


    def test_read_fid(self):

        fid = glob.glob(vj.config['fids_dir']+'/gems*')[0]
        varr = vj.read_fid(fid)
        self.assertEqual(varr.source,'fid')
        self.assertEqual(len(varr.data.shape),2)
        self.assertEqual(varr.space,'fid')
        self.assertEqual(varr.pd['pslabel'],'gems')
        self.assertEqual(varr.apptype,'im2D')


    def test_read_fdf(self):

        fdf = glob.glob(vj.config['fdfs_dir']+'/gems*')[0]
        varr = vj.read_fdf(fdf)
        self.assertEqual(varr.source,'fdf')
        self.assertEqual(len(varr.data.shape),4)
        self.assertEqual(varr.pd['pslabel'],'gems')
        self.assertEqual(varr.apptype,'im2D')

    def test_read_nifti(self):

        pass



import numpy as np
from lmfit import minimize, Parameters
import matplotlib.pyplot as plt
import itertools
import vnmrjpy as vj

"""
Collection of functions aiding the fitting of 3D timeseries data
"""

def residual1D(params, model, x, data, eps_data):
    """Return residual for least squares fitting"""
    return (data-model) / eps_data

def minimize3D(residual, params3D, args3D, mask3D):
    """Wrapper for lmfit.minimize for 3D timeseries

    Intended behavior is the same, except inputs and outputs are in
    numpy arrays.

    Args:
        residual -- residual function for 1D data
        params3D -- lmfit.Parameters in 3D numpy array
        args3D -- (x,data, ydata, eps_data) in 3D array
        mask3D -- 3D np.array of same spatial shape as data
    Return:
        out3D
    """
    (x_dim, y_dim, z_dim) = args3D[1].shape[0:3]
    out3D = np.zeros((x_dim,y_dim, z_dim))
    for x, y, z in itertools.product(*map(range, (x_dim,y_dim,z_dim))):
        if mask3D[x,y,z] == 0:
            zero = minimize.MinimizerResults()
            out3D[x,y,z] = 1        
            continue
        else:
            (xdata,ydata,eps_data) = args3D[x,y,z]
            params = params3D[x,y,z]
            try:
                out3D[x,y,z] =\
                 minimize(residual, params, args=(xdata,ydata, eps_data))
            except:
                pass
    return out3D

def fit3D_leastsq_2stage():

    pass

def reinit3D():

    pass




import vnmrjpy as vj
import numpy as np
from functools import reduce
from vnmrjpy.core.utils import vprint
import warnings
import matplotlib.pyplot as plt

"""
vnmrjpy.varray
==============

Contaions the main varray object, and the basic k-space and image space
building methods as well as conversion between the two.

"""
class varray():
    """Main vnmrjpy object to carry data and reconstruction information.

    Attributes:
        
    Methods:


    """
    def __init__(self, data=None, pd=None, source=None, is_zerofilled=False,\
                is_kspace_complete=False, fid_header=None, fdf_header=None,\
                nifti_header=None, space=None, intent=None,dtype=None,\
                arrayed_params=[None,1], seqcon=None, apptype=None,\
                vdtype=None, sdims=None, dims=None, description=None):

        self.data = data
        self.pd = pd
        self.source = source
        self.intent = intent
        self.space = space
        self.dtype= dtype
        self.vdtype = vdtype
        self.is_zerofilled = is_zerofilled
        self.is_kspace_complete = is_kspace_complete
        self.fid_header = fid_header
        self.fdf_header = fid_header
        self.nifti_header = nifti_header
        self.arrayed_params = arrayed_params
        self.seqcon = seqcon
        self.apptype = apptype
        self.sdims = sdims
        self.dims = dims
        self.description = description
    
    def flip_axis(self,axis):
        """Flip data on axis 'x','y','z' or 'phase','read','slice'"""
        return vj.core.transform._flip_axis(self,axis)        

    def set_nifti_header(self):

        return vj.core.niftitools._set_nifti_header(self)

    def set_dims(self):
        """Set dims, meaning x,y,z axes order""" 
        return vj.core.transform._set_dims(self)

    # put the transforms into another file for better visibility

    def to_local(self):
        """Transform to original space at k-space formation"""
        return vj.core.transform._to_local(self)

    def to_scanner(self):
        """Transform data to scanner coordinate space by properly swapping axes

        Standard vnmrj orientation - meaning rotations are 0,0,0 - is axial, with
        x,y,z axes (as global gradient axes) corresponding to phase, readout, slice.
        vnmrjpy defaults to handling numpy arrays of:
                    (readout, phase, slice/phase2, time*echo, receivers)
        but arrays of
                            (x,y,z, time*echo, reveivers)
        is also desirable in some cases (for example registration in FSL flirt,
        or third party stuff)

        Euler angles of rotations are psi, phi, theta.

        Also corrects reversed X gradient and sliceorder
        """
        if vj.core.transform._check_90deg(self.pd):
            pass
        else:
            raise(Exception('Only Euler angles of 0,90,180 are permitted.'))    

        return vj.core.transform._to_scanner(self)

    def to_anatomical(self):
        """Transform to rat brain anatomical coordinate system, the 'usual'"""
        return vj.core.transform._to_anatomical(self)

    def to_global(self):
        """Fixed to scanner, but enlarges FOV if necessary to support oblique.
        """
        return vj.core.transform._to_global(self)

    def to_kspace(self):
        """Build the k-space from the raw fid data and procpar.

        Raw fid_data is numpy.ndarray(blocks, traces * np) format. Should be
        untangled based on 'seqcon' or 'seqfil' parameters.
        seqcon chars refer to (echo, slice, Pe1, Pe2, Pe3)

        PREVIOUS:
                ([rcvrs, phase, read, slice, echo*time])
        NOW:
                ([phase, read, slice, echo*time, rcvrs])
        """
        # if data is not from fid, just fft it
        if self.vdtype == 'imagespace':
            raise(Exception('not implemented yet'))        
            #TODO something like this:

            #self.data = vj.core.recon._fft(self.data,dims)
            #self.vdype = 'kspace'                

        # check if data is really from fid
        if self.vdtype is not 'fid':
            raise(Exception('varray data is not fid data.'))
        self.data = np.vectorize(complex)(self.data[:,0::2],\
                                                self.data[:,1::2])
        # check for arrayed parameters, save the length for later 
        array_length = reduce(lambda x,y: x*y, \
                        [i[1] for i in self.arrayed_params])
        blocks = self.data.shape[0] // array_length
        vprint('Making k-space for '+ str(self.apptype)+' '\
                +str(self.pd['seqfil'])+' seqcon: '+str(self.pd['seqcon']))
        rcvrs = self.pd['rcvrs'].count('y')

        def _is_interleaved(ppdict):
            res  = (int(ppdict['sliceorder']) == 1)
            return res
        def _is_evenslices(ppdict):
            try:
                res = (int(ppdict['ns']) % 2 == 0)
            except:
                res = (int(ppdict['pss']) % 2 == 0)
            return res
        def make_im2D():
            """Child method of 'make', provides the same as vnmrj im2Drecon"""
            p = self.pd 
            rcvrs = int(p['rcvrs'].count('y'))
            (read, phase, slices) = (int(p['np'])//2,int(p['nv']),int(p['ns']))
            if 'ne' in p.keys():
                echo = int(p['ne'])
            else:
                echo = 1
            time = 1
            # this is the old shape which worked, better to reshape at the end
            finalshape = (rcvrs, phase, read, slices,echo*time*array_length)
            final_kspace = np.zeros(finalshape,dtype='complex64')
           
            for i in range(array_length):
 
                kspace = self.data[i*blocks:(i+1)*blocks,...]

                if p['seqcon'] == 'nccnn':
                    shape = (rcvrs, phase, slices, echo*time, read)
                    kspace = np.reshape(kspace, shape, order='C')
                    kspace = np.moveaxis(kspace, [0,1,4,2,3], [0,1,2,3,4])
                    

                elif p['seqcon'] == 'nscnn':

                    raise(Exception('not implemented'))

                elif p['seqcon'] == 'ncsnn':

                    preshape = (rcvrs, phase, slices*echo*time*read)
                    shape = (rcvrs, phase, slices, echo*time, read)
                    kspace = np.reshape(kspace, preshape, order='F')
                    kspace = np.reshape(kspace, shape, order='C')
                    kspace = np.moveaxis(kspace, [0,1,4,2,3], [0,1,2,3,4])

                elif p['seqcon'] == 'ccsnn':

                    preshape = (rcvrs, phase, slices*echo*time*read)
                    shape = (rcvrs, phase, slices, echo*time, read)
                    kspace = np.reshape(kspace, preshape, order='F')
                    kspace = np.reshape(kspace, shape, order='C')
                    kspace = np.moveaxis(kspace, [0,1,4,2,3], [0,1,2,3,4])
                else:
                    raise(Exception('Not implemented yet'))
                if _is_interleaved(p): # 1 if interleaved slices
                    if _is_evenslices(p):
                        c = np.zeros(kspace.shape, dtype='complex64')
                        c[...,0::2,:] = kspace[...,:slices//2,:]
                        c[...,1::2,:] = kspace[...,slices//2:,:]
                        kspace = c
                    else:
                        c = np.zeros(kspace.shape, dtype='complex64')
                        c[...,0::2,:] = kspace[...,:(slices+1)//2,:]
                        c[...,1::2,:] = kspace[...,(slices-1)//2+1:,:]
                        kspace = c

                final_kspace[...,i*echo*time:(i+1)*echo*time] = kspace
            self.data = final_kspace
            return self

        def make_im2Dcs():
            """
            These (*cs) are compressed sensing variants
            """

            def decode_skipint_2D(skipint):

               pass
 
            raise(Exception('not implemented'))

        def make_im2Depi():

            p = self.pd
            
            comp_seg = p['cseg']
            altread = p['altread']
            nseg = int(p['nseg'])
            etl = int(p['etl'])
            kzero = int(p['kzero'])  
            images = int(p['images'])  # repetitions
            rcvrs = int(p['rcvrs'].count('y'))
            echo = 1

            if p['navigator'] == 'y':
                pluspe = 1 + int(p['nnav'])  # navigator echo + unused
            else:
                pluspe = 1  # unused only
            
            print('images {}'.format(images))
            print('nseg {}'.format(nseg))
            print('ns {}'.format(p['ns']))
            print('seqcon {}'.format(p['seqcon']))

            # number of acquisitions based on epi ref.  
            # see manual for explanation
            epiref = p['epiref_type']
            if epiref ==  'fulltriple':
                time = 2 + 2*images
            elif epiref == 'triple':
                time = 3+images
            elif epiref == 'single':
                time = 1+images
            elif epiref == 'none':
                time = images
            else:
                raise(Exception('This type of epiref is not implemented'))
            if int(p['pro']) != 0:
                print('pro not 0')
                (read, phase, slices) = (int(p['nread']), \
                                            int(p['nphase']), \
                                            int(p['ns']))
            else:
                (read, phase, slices) = (int(p['nread'])//2, \
                                            int(p['nphase']), \
                                            int(p['ns']))
            
            print('read {}, phase {}, slices {}'.format(read,phase, slices))
            finalshape = (rcvrs, phase, read, slices,echo*time*array_length)
            final_kspace = np.zeros(finalshape,dtype='complex64')
            print('fid data shape {}'.format(self.data.shape))
            print('blocks {}'.format(blocks))

            for i in range(array_length):
 
                prekspace = self.data[i*blocks:(i+1)*blocks,...]
                print('kspace shape in outer {}'.format(prekspace.shape))

                # this case repetiotions are in different blocks
                if p['seqcon'] == 'ncnnn':
                
                    #kspace = np.zeros((rcvrs,etl+pluspe,slices,time,read),\
                    #                    dtype='complex64')
                    kspace = np.zeros((rcvrs,slices,etl+pluspe,time,read),\
                                        dtype='complex64')
                    for t in range(time):

                        kspace_t = prekspace[t*rcvrs:(t+1)*rcvrs,...]
                        preshape = (rcvrs, slices, etl+pluspe, 1, read)
                        kspace_t = np.reshape(kspace_t, preshape, order='c')
                        
                        kspace[...,t:t+1,:] = kspace_t
                        # TODO
                        # get navigator echoes out

                        # TODO
                        # navigator correct
                    
                    kspace = np.moveaxis(kspace,[0,1,2,3,4],[0,3,1,4,2])
                    print('kspace shape HERE {}'.format(kspace.shape))
                else:
                    raise(Exception('This seqcon not implemented in epip'))

                if _is_interleaved(p): # 1 if interleaved slices
                    if _is_evenslices(p):
                        c = np.zeros(kspace.shape, dtype='complex64')
                        c[...,0::2,:] = kspace[...,:slices//2,:]
                        c[...,1::2,:] = kspace[...,slices//2:,:]
                        kspace = c
                    else:
                        c = np.zeros(kspace.shape, dtype='complex64')
                        c[...,0::2,:] = kspace[...,:(slices+1)//2,:]
                        c[...,1::2,:] = kspace[...,(slices-1)//2+1:,:]
                        kspace = c

                # -------------------epi kspace processing---------------------

                # get navigator echo out of data
                navigators = vj.core.epitools._get_navigator_echo(kspace,p)
                # remove unused echo and navigator, fill rest up with 0
                kspace = vj.core.epitools._prepare_shape(kspace,p)
                #kspace = vj.core.epitools._kzero_shift(kspace,p)
                kspace = vj.core.epitools._reverse_even(kspace)

                final_kspace[...,i*echo*time:(i+1)*echo*time] = kspace
                #TODO this is only for testing
                img = final_kspace[1,:,:,10,3]
                fft = np.fft.fftshift(img)
                fft = np.fft.fft2(fft)
                fft = np.fft.ifftshift(fft)
                plt.subplot(1,2,1)
                plt.imshow(np.absolute(img))
                plt.subplot(1,2,2)
                plt.imshow(np.absolute(fft))
                plt.show()

            self.data = final_kspace
            return self

        def make_im2Depics():
            raise(Exception('not implemented'))
        def make_im2Dfse():

            p = self.pd
            #petab = vj.util.getpetab(self.procpar,is_procpar=True)
            petab = vj.core.read_petab(self.pd)
            nseg = int(p['nseg'])  # seqgments
            etl = int(p['etl'])  # echo train length
            kzero = int(p['kzero'])  
            images = int(p['images'])  # repetitions
            (read, phase, slices) = (int(p['np'])//2,int(p['nv']),int(p['ns']))

            # setting time params
            echo = 1
            time = images

            phase_sort_order = np.reshape(np.array(petab),petab.size,order='C')
            # shift to positive
            phase_sort_order = phase_sort_order + phase_sort_order.size//2-1

            finalshape = (rcvrs, phase, read, slices,echo*time*array_length)
            final_kspace = np.zeros(finalshape,dtype='complex64')
           
            for i in range(array_length):
 
                kspace = self.data[i*blocks:(i+1)*blocks,...]
                if p['seqcon'] == 'nccnn':

                    #TODO check for images > 1
                    preshape = (rcvrs, phase//etl, slices, echo*time, etl, read)
                    shape = (rcvrs, echo*time, slices, phase, read)
                    kspace = np.reshape(kspace, preshape, order='C')
                    kspace = np.swapaxes(kspace,1,3)
                    kspace = np.reshape(kspace, shape, order='C')
                    # shape is [rcvrs, phase, slices, echo*time, read]
                    kspace = np.swapaxes(kspace,1,3)
                    kspace_fin = np.zeros_like(kspace)
                    kspace_fin[:,phase_sort_order,:,:,:] = kspace
                    kspace_fin = np.moveaxis(kspace_fin, [0,1,4,2,3], [0,1,2,3,4])
                    kspace = kspace_fin
                else:
                    raise(Exception('not implemented'))
                
                if _is_interleaved(p): # 1 if interleaved slices
                    if _is_evenslices(p):
                        c = np.zeros(kspace.shape, dtype='complex64')
                        c[...,0::2,:] = kspace[...,:slices//2,:]
                        c[...,1::2,:] = kspace[...,slices//2:,:]
                        kspace = c
                    else:
                        c = np.zeros(kspace.shape, dtype='complex64')
                        c[...,0::2,:] = kspace[...,:(slices+1)//2,:]
                        c[...,1::2,:] = kspace[...,(slices-1)//2+1:,:]
                        kspace = c
            
                final_kspace[...,i*echo*time:(i+1)*echo*time] = kspace
            self.data = final_kspace
            return self

        def make_im2Dfsecs():
            raise(Exception('not implemented'))
        def make_im3D():
            """Child method of 'make', provides the same as vnmrj im3Drecon"""
            p = self.pd 
            rcvrs = int(p['rcvrs'].count('y'))
            (read, phase, phase2) = (int(p['np'])//2,int(p['nv']),int(p['nv2']))
            if 'ne' in p.keys():
                echo = int(p['ne'])
            else:
                echo = 1
            if 'images' in p.keys():
                time = int(p['images'])
            else:
                time = 1

            finalshape = (rcvrs, phase, read, phase2,echo*time*array_length)
            final_kspace = np.zeros(finalshape,dtype='complex64')
           
            for i in range(array_length):
 
                kspace = self.data[i*blocks:(i+1)*blocks,...]

                if p['seqcon'] == 'nccsn':
                
                    preshape = (rcvrs,phase2,phase*echo*time*read)
                    shape = (rcvrs,phase2,phase,echo*time,read)
                    kspace = np.reshape(kspace,preshape,order='F')
                    kspace = np.reshape(kspace,shape,order='C')
                    kspace = np.moveaxis(kspace, [0,2,4,1,3], [0,1,2,3,4])
                    # what is this??
                    #kspace = np.flip(kspace,axis=3)

                if p['seqcon'] == 'ncccn':
                    preshape = (rcvrs,phase2,phase*echo*time*read)
                    shape = (rcvrs,phase,phase2,echo*time,read)
                    kspace = np.reshape(kspace,preshape,order='F')
                    kspace = np.reshape(kspace,shape,order='C')
                    kspace = np.moveaxis(kspace, [0,2,4,1,3], [0,1,2,3,4])
        
                if p['seqcon'] == 'cccsn':
                
                    preshape = (rcvrs,phase2,phase*echo*time*read)
                    shape = (rcvrs,phase,phase2,echo*time,read)
                    kspace = np.reshape(kspace,preshape,order='F')
                    kspace = np.reshape(kspace,shape,order='C')
                    kspace = np.moveaxis(kspace, [0,2,4,1,3], [0,1,2,3,4])

                if p['seqcon'] == 'ccccn':
                    
                    shape = (rcvrs,phase2,phase,echo*time,read)
                    kspace = np.reshape(kspace,shape,order='C')
                    kspace = np.moveaxis(kspace, [0,2,4,1,3], [0,1,2,3,4])

                final_kspace[...,i*echo*time:(i+1)*echo*time] = kspace

            self.data = final_kspace
            
            return self

        def make_im3Dcs():
            """
            3D compressed sensing
            sequences : ge3d, mge3d, se3d, etc
            """
            # -------------------im3Dcs Make helper functions ---------------------

            def decode_skipint_3D(skipint):
                """
                Takes 'skipint' parameter and returns a 0-1 matrix according to it
                which tells what lines are acquired in the phase1-phase2 plane
                """
                BITS = 32  # Skipint parameter is 32 bit encoded binary, see spinsights
                skip_matrix = np.zeros([int(p['nv']), int(p['nv2'])])
                skipint = [int(x) for x in skipint]
                skipint_bin_vals = [str(np.binary_repr(d, BITS)) for d in skipint]
                skipint_bin_vals = ''.join(skipint_bin_vals)
                skipint_bin_array = np.asarray([int(i) for i in skipint_bin_vals])
                skip_matrix = np.reshape(skipint_bin_array, skip_matrix.shape)

                return skip_matrix

            def fill_kspace_3D(pre_kspace, skip_matrix, shape):
                """
                Fills up reduced kspace with zeros according to skip_matrix
                returns zerofilled kspace in the final shape
                """
                kspace = np.zeros(shape, dtype=complex)
                if self.p['seqcon'] == 'ncccn':

                    n = int(self.p['nv'])
                    count = 0
                    for i in range(skip_matrix.shape[0]):
                        for k in range(skip_matrix.shape[1]):
                            if skip_matrix[i,k] == 1:
                                kspace[:,i,k,:,:] = pre_kspace[:,count,:,:]
                                count = count+1
                self.kspace = kspace
                return kspace

            #------------------------im3Dcs make start -------------------------------

            kspace = self.pre_kspace
            p = self.p
            (read, phase, phase2) = (int(p['np'])//2, \
                                    int(p['nv']), \
                                     int(p['nv2']))

            shiftaxis = (self.config['pe_dim'],\
                        self.config['ro_dim'],\
                        self.config['pe2_dim'])

            if 'ne' in p.keys():
                echo = int(p['ne'])
            else:
                echo = 1

            time = 1

            if p['seqcon'] == 'nccsn':

                pass

            if p['seqcon'] == 'ncccn':
            
                skip_matrix = decode_skipint_3D(p['skipint'])
                pre_phase = int(self.fid_header['ntraces'])    
                shape = (self.rcvrs, phase, phase2, echo*time, read)
                pre_shape = (self.rcvrs, pre_phase, echo*time, read)
                pre_kspace = np.reshape(kspace, pre_shape, order='c')
                kspace = fill_kspace_3D(pre_kspace, skip_matrix, shape)
                kspace = np.moveaxis(kspace, [0,1,4,2,3],[0,1,2,3,4])
                
            self.kspace = kspace
            return kspace

        def make_im3Dute():
            raise(Exception('not implemented')) 
        # ----------------Handle sequence exceptions first---------------------

        vprint(' making seqfil : {}'.format(self.pd['seqfil']))


        if str(self.pd['seqfil']) == 'ge3d_elliptical':
           
            self = make_im3Dcs()

        #--------------------------Handle by apptype---------------------------

        if self.pd['apptype'] == 'im2D':

            self = make_im2D()
            self.is_kspace_complete = True

        elif self.pd['apptype'] == 'im2Dcs':

            self = make_im2Dcs()

        elif self.pd['apptype'] == 'im2Depi':

            self = make_im2Depi()
            self.is_kspace_complete = True

        elif self.pd['apptype'] == 'im2Depics':

            self = make_im2Depics()

        elif self.pd['apptype'] == 'im2Dfse':

            self = make_im2Dfse()
            self.is_kspace_complete = True

        elif self.pd['apptype'] == 'im2Dfsecs':

            self = make_im2Dfsecs()

        elif self.pd['apptype'] == 'im3D':

            self = make_im3D()
            self.is_kspace_complete = True

        elif self.pd['apptype'] == 'im3Dcs':

            self = make_im3Dcs()

        elif self.pd['apptype'] == 'im3Dute':

            self = make_im3Dute()
            self.is_kspace_complete = True

        else:
            raise(Exception('Could not find apptype. Maybe not implemented?'))

        # ---------------------Global modifications on Kspace------------------

        #TODO if slices are in reversed order, flip them
        
        # in revamp make new axes, mainly for nifti io, and viewers
        # old : [rcvrs, phase, read, slice, time]
        # new : [read, phase, slice, time, rcvrs]

        self.data = np.moveaxis(self.data,[0,1,2,3,4],[4,1,0,2,3])
        # swap axes 0 and 1 so phase, readout etc is the final order
        self.data = np.swapaxes(self.data,0,1)
        self.vdtype='kspace'
        self.to_local()

        if vj.config['default_space'] == None:
            pass
        elif vj.config['default_space'] == 'anatomical':
            self.to_anatomical()
        
        return self

    def to_imagespace(self):
        """ Reconstruct MR images to real space from k-space.

        Generally this is done by fourier transform and corrections.
        Hardcoded for each 'seqfil' sequence.

        Args:

        Updates attributes:
            data
            

        """
        seqfil = str(self.pd['seqfil'])
        # this is for fftshift
        ro_dim = self.sdims.index('read')  # this should be default
        pe_dim = self.sdims.index('phase')  # this should be default
        pe2_dim = self.sdims.index('slice')  # slice dim is also pe2 dim
        
        sa = (ro_dim, pe_dim, pe2_dim)

        if seqfil in ['gems', 'fsems', 'mems', 'sems', 'mgems']:

            self.data = vj.core.recon._ifft(self.data,sa[0:2])

        elif seqfil in ['ge3d','fsems3d','mge3d']:
            
            self.data = vj.core.recon._ifft(self.data,sa)

        elif seqfil in ['ge3d_elliptical']:

            self.data = vj.core.recon._ifft(self.data,sa)

        else:
            raise Exception('Sequence reconstruction not implemented yet')
        
        self.vdtype = 'imagespace'

        return self




import vnmrjpy as vj
import unittest
import glob

class Test_Composer(unittest.TestCase):

    def test_composer(self):

        compdir = vj.config['dataset_dir']+'/parameterfit/composer'
        ref = glob.glob(compdir+'/*composer08_01.fid')[0]
        mems = glob.glob(compdir+'/mems_s_2019022301_01*')[0]

        varr = vj.core.read_fid(mems).to_kspace().to_imagespace()
        varr.to_anatomical()
        varr_ref = vj.core.read_fid(ref).to_kspace().to_imagespace()
        varr_ref.to_anatomical()

        test_workdir = vj.config['fsl_workdir']
        varr_matched = vj.func.COMPOSER(varr,varr_ref,\
                            workdir=None,keepfiles=False)
        self.assertEqual(varr_matched.data.shape,varr.data.shape)



import vnmrjpy as vj
import unittest
import numpy as np
import glob
import matplotlib.pyplot as plt

class Test_base_mask(unittest.TestCase):

    def test_mask(self):

        seq = glob.glob(vj.config['fids_dir']+'/gems*axial_*')[0]
        print(seq)
        varr = vj.core.read_fid(seq).to_kspace().to_imagespace()
        varr.to_anatomical()
        print(varr.data.shape)
        magn = vj.core.recon.ssos(varr.data)
        print(magn.shape)
        print(varr.sdims)
        print(varr.dims)
        mask = vj.func.base.mask(varr)
        print('Mask should be made with BET. Consider this again....')



import vnmrjpy as vj
import unittest
import glob
import os

class Test_CsTestDataGenerator(unittest.TestCase):

    def test_testdatagen_gems(self):

        # directories
        fid_dir = sorted(glob.glob(vj.fids+'/gems_s*'))[0]
        base_dir = os.path.basename(fid_dir)[:-4]
        red = 4
        out_dir = vj.cs+'/'+base_dir+'_red'+str(red)+'.cs'
        # filepaths
        fid = fid_dir+'/fid'
        procpar = fid_dir+'/procpar'
        # generation
        gen = vj.util.CsTestDataGenerator(fid,procpar,reduction=red)
        gen.generate(savedir=out_dir)

    def test_testdatagen_angio(self):

        # directories
        fid_dir = sorted(glob.glob(vj.fids+'/ge3d_angio*'))[0]
        base_dir = os.path.basename(fid_dir)[:-4]
        #cs reduction
        red = 8
        out_dir = vj.cs+'/'+base_dir+'_red'+str(red)+'.cs'
        # filepaths
        fid = fid_dir+'/fid'
        procpar = fid_dir+'/procpar'
        # generation
        gen = vj.util.CsTestDataGenerator(fid,procpar,reduction=red)
        gen.generate(savedir=out_dir)

    def test_testdatagen_ge3d(self):

        # directories
        fid_dir = sorted(glob.glob(vj.fids+'/ge3d_s*'))[0]
        base_dir = os.path.basename(fid_dir)[:-4]
        red = 4
        out_dir = vj.cs+'/'+base_dir+'_red'+str(red)+'.cs'
        # filepaths
        fid = fid_dir+'/fid'
        procpar = fid_dir+'/procpar'
        # generation
        gen = vj.util.CsTestDataGenerator(fid,procpar,reduction=red)
        gen.generate(savedir=out_dir)

    def test_testdatagen_mems(self):

        # directories
        fid_dir = sorted(glob.glob(vj.fids+'/mems*'))[0]
        base_dir = os.path.basename(fid_dir)[:-4]
        red = 6
        out_dir = vj.cs+'/'+base_dir+'_red'+str(red)+'.cs'
        # filepaths
        fid = fid_dir+'/fid'
        procpar = fid_dir+'/procpar'
        # generation
        gen = vj.util.CsTestDataGenerator(fid,procpar,reduction=red)
        gen.generate(savedir=out_dir)



import vnmrjpy as vj
import numpy as np
import unittest
import nibabel as nib
import os

class Test_T2fitter(unittest.TestCase):

    def test_t2fit(self):

        nifti = vj.niftis+'/mems/mems_20180523_01.nii'
        try:
            procpar = vj.niftis+'/mems/.mems_20180523_01.procpar'
        except:
            procpar = glob.glob(vj.niftis+'/mems/*mems_20180523*')[0]
        # init for output saving
        test_out_dir = vj.config['testresults_dir']+'/fit/t2/'
        if not os.path.exists(test_out_dir):
            os.makedirs(test_out_dir)
        name = 't2map_mems_20180523'
        full_out = test_out_dir+name

        data = nib.load(nifti).get_fdata()
        ppdict = vj.io.ProcparReader(procpar).read()

        fitter = vj.fit.T2Fitter(data, procpar=procpar, automask=True, fitfunc='3param_exp')
        pars = fitter.fit()
        print(len(pars))
        t2map, m0map, const = pars
        print(t2map.shape)
        writer = vj.io.NiftiWriter(t2map,procpar)
        writer.write(full_out)




#!/usr/bin/python3

import numpy as np
import matplotlib.pyplot as plt
import time

class RealTimeImshow():
    """
    Utility class for checking matrix iterations
    INPUT: initial 2D matrix

    METHODS:
        update_data(data) : replaces plot data
    """
    def __init__(self,init_data,cmap='gray',sleeptime=0.01):

        plt.figure()
        self.sleeptime = sleeptime
        self.ax = plt.subplot(1,1,1)
        self.img = plt.imshow(init_data,cmap=cmap)
        plt.show(block=False)
        plt.pause(0.001)

    def update_data(self,data):

        self.img.set_data(data)
        plt.pause(0.001)
        time.sleep(self.sleeptime)

# -------------------------TESTING----------------------------------------------

def main():

    def generate_data_():

        data = np.random.rand(100,100)
        return data


    init_data = generate_data_()
    rtplot = RealTimeImshow(init_data)
    for _ in range(100):
        data = generate_data_()
        rtplot.update_data(data)
        time.sleep(0.01)

if __name__ == '__main__':
    main()



import copy
import numpy as np
import sys
import vnmrjpy as vj
import time

class Admm():
    """Alternating Direction Method of Multipliers solver for Aloha


    Tuned for ALOHA MRI reconstruction framework, not for general use yet.
    Lmafit estimates the rank, then Admm is used to enforce the hankel structure

    refs.: Aloha papers ?
            Admm paper ?
    """
    def __init__(self,U,V,fiber_stage_known,stage,rp,\
                mu=1000,\
                realtimeplot=False,\
                noiseless=True,\
                device='CPU'):
        """Initialize solver
        
        Args:
            U, V  (np.matrix) U*V.H is the estimated hankel matrix from Lmafit
            slice3d_cs_weighted : 
            slice3d_shape : 
            stage (int) : pyramidal decomposition stage ?? WHY??
            rp {dictionary} : Aloha recon parameters 
            realitmeplot (Boolean) option to plot each iteration

        """
        #TODO change the old hankel compose-decompose to the new one
        fiber_shape = rp['fiber_shape']
        #removed np matrix
        # a.H syntax is changed to a.conj().T
        #U = np.matrix(U)
        #V = np.matrix(V)
        # make ankel mask out of known elmenets
        hankel_mask = np.absolute(vj.aloha.construct_hankel(fiber_stage_known,rp))
        hankel_mask[hankel_mask != 0] = 1
        hankel_mask = np.array(hankel_mask,dtype='complex64')
        hankel_mask_inv = np.ones(hankel_mask.shape) - hankel_mask

        self.hankel_mask = hankel_mask
        self.hankel_mask_inv = hankel_mask_inv
        # real time plotting for debugging purposes
        self.realtimeplot = realtimeplot
        if realtimeplot == True:
            self.rtplot = vj.util.RealTimeImshow(np.absolute(U.dot(V.conj().T)))

        # putting initpars into tuple
        self.initpars = (U,V,fiber_stage_known,stage,rp,mu,noiseless)

    def solve(self,max_iter=100):
        """The actual Admm iteration.
        
        Returns:
            hankel = U.dot(V.H) (np.matrix)
        """
        (U,V,fiber_stage,s,rp,mu,noiseless) = self.initpars

        hankel = U.dot(V.conj().T)

        fiber_orig_part = copy.deepcopy(fiber_stage)
        # init lagrangian update
        lagr = np.zeros(hankel.shape,dtype='complex64')
        #lagr = copy.deepcopy(hankel)
        us = (U.conj().T.dot(U)).shape
        vs = (V.conj().T.dot(V)).shape
        Iu = np.eye(us[0],us[1],dtype='complex64')
        Iv = np.eye(vs[0],vs[0],dtype='complex64')


        for _ in range(max_iter):
        
            #start = time.time()
            # taking the averages from tha hankel structure and rebuild
            hankel_inferred_part = np.multiply(\
                                U.dot(V.conj().T)-lagr,self.hankel_mask_inv)  
            #dtime = time.time()
            fiber_inferred_part = vj.aloha.deconstruct_hankel(\
                                hankel_inferred_part,s,rp)
            #print('deconstruct time {}'.format(time.time()-dtime))
            fiber = fiber_orig_part + fiber_inferred_part
            hankel = vj.aloha.construct_hankel(fiber,rp)
            # updating U,V and the lagrangian
            #TODO consider multidot....
            U = mu*(hankel+lagr).dot(V).dot(\
                                np.linalg.inv(Iv+mu*V.conj().T.dot(V)))
            V = mu*((hankel+lagr).conj().T).dot(U).dot(\
                                np.linalg.inv(Iu+mu*U.conj().T.dot(U)))
            lagr = hankel - U.dot(V.conj().T) + lagr

            if self.realtimeplot == True:
                self.rtplot.update_data(np.absolute(U.dot(V.conj().T)))
            #print('admm iter time {}'.format(time.time()-start))

        return U.dot(V.conj().T)
    



"""
Vnmrjpy utility functions
=========================


"""

from .plotutils import *
from .configparser import *
from .csdatagen import *
from .skipinttools import *
from .cstestio import *
from .readpetable import *
from .transform import *
from .disgusting import *
from .niftiutils import *
from .fidutils import *



import unittest
import vnmrjpy as vj
import nibabel as nib
import numpy as np
import glob

class Test_nifti(unittest.TestCase):

    def test_gems(self):

        vj.config['verbose'] = False
        seq = glob.glob(vj.config['fids_dir']+'/gems*axial_*')[0]
        varr = vj.core.read_fid(seq)
        
        varr.to_kspace().to_imagespace()
        #varr = vj.core.niftitools._set_nifti_header(varr)
        niftiout = vj.config['testresults_dir']+'/new'
        niftiout = niftiout + '/testnifti'
        vj.core.write.write_nifti(varr,niftiout) 


        



import os
import glob
import numpy as np
import nibabel as nib
import vnmrjpy as vj



class FdfReader():
    """Read Vnmrj .fdf files

    data = np.ndarray([receivers, phase, readout, slice, time])    


    """

    def __init__(self, inPath, procpar=None, printlines=False, verbose=False):
        self.path = inPath
        self.procpar = procpar
        self.printlines = printlines
        self.conf = vj.config
    
    def read(self):
        """Read file

        Returns: (header, data)
        """        
        #---------------- auxiliary functions for read method -----------------

        def preproc_fdf(fdf):

            with open(fdf,'rb') as openFdf:
                fdata = bytearray(openFdf.read())
                nul = fdata.find(b'\x00')
                header = fdata[:nul]
                data = fdata[nul+1:]
            return (header,data)

        # ----------parse fdf header and return into a dictionary --------------

        def parse_header(header):
            keys_to_parse = sorted(['rank','roi','location','spatial_rank',\
                            'matrix','orientation',\
                            'studyid','gap','pe_size','ro_size',\
                            'pe2_size', 'abscissa',\
                            'storage'])
            to_delete = ('char','float','int')
            header = header.decode('ascii').split('\n')
            header_dict = {}    
            for line in header:             # some formatting of header
                if self.printlines:
                    print(line)
                for item in to_delete:
                    if line.startswith(item):
                        line = line.split(item,1)
                        break
                try:
                    line = line[1].lstrip()
                    line = line.lstrip('*').rstrip(';')
                    if '[]' in line:
                        line = line.replace('[]','')
                    if '{' in line:
                        line = line.replace('{','(')
                    if '}' in line:
                        line = line.replace('}',')')
                    if ' ' in line:
                        line = line.replace(' ','')
                    line = line.split('=')
                    header_dict[line[0]] = line[1]        
                except:
                    continue

            for item in keys_to_parse:
                if item in header_dict.keys():            
                    if item == 'abscissa':
                        tempval = header_dict[item][1:-1];''.join(tempval)
                        tempval = tempval.replace('"','')                
                        header_dict[item] = tuple([k \
                                        for k in tempval.split(',')])
                    if item == 'matrix':            
                        tempval = header_dict[item][1:-1];''.join(tempval)                
                        header_dict[item] = tuple([int(k) \
                                        for k in tempval.split(',')])
                    if item == 'roi':
                        tempval = header_dict[item][1:-1];''.join(tempval)                
                        header_dict[item] = tuple([float(k)\
                                        for k in tempval.split(',')])
                    if item == 'ro_size' or item == 'pe_size' \
                                        or item == 'pe2_size':
                        header_dict[item] = int(header_dict[item])
                    if item == 'storage':
                        tempval = header_dict[item];''.join(tempval)
                        tempval = tempval.replace('"','')
                        header_dict[item] = str(tempval)
                    if item == 'orientation':
                        tempval = header_dict[item][1:-1];''.join(tempval)                
                        header_dict[item] = tuple([float(k)\
                                        for k in tempval.split(',')])
                    if item == 'location':
                        tempval = header_dict[item][1:-1];''.join(tempval)                
                        header_dict[item] = tuple([float(k)\
                                        for k in tempval.split(',')])
                    if item == 'gap':
                        header_dict[item] = float(header_dict[item])
                    if item == 'slices':
                        header_dict[item] = int(header_dict[item])
                    if item == 'TR':
                        header_dict[item] = float(header_dict[item])/1000
            return header_dict

        #----------process bynary data based on header--------------------
    
        def prepare_data(binary_data):

            matrix = self.header_dict['matrix']

            if self.header_dict['storage'] == 'float' and \
               self.header_dict['bits'] == '32':
                dt = np.dtype('float32'); dt = dt.newbyteorder('<')

            else:
                print('')
                print('error: data type incorrectly specified in "prepare_data"\n')
                return -1

            img_data = np.frombuffer(binary_data, dtype=dt)
            img_data = np.reshape(img_data,matrix)
            return img_data
        #---------------------------------------------------------------------
        #                             main read method
        #----------------------------------------------------------------------
        if os.path.isdir(self.path):
            self.procpar = str(self.path)+'/procpar'
            fdf_list = sorted(glob.glob(str(self.path)+'/*.fdf'))
            # run preproc once tp get the header
            (header, data) = preproc_fdf(fdf_list[0])
            ppr = vj.io.ProcparReader(self.procpar)
            self.ppdict = ppr.read()
        else:
            try:            
                ppr = vj.io.ProcparReader(self.procpar)
                self.ppdict = ppr.read()
            except:
                pass
                #print('\nFdfReader.read() warning : Please specify procpar file!\n')
            fdf_list =[self.path]
            (header, data) = preproc_fdf(self.path)
            
        self.header_dict = parse_header(header)
        # ------------------------process if 3d --------------------
        if self.header_dict['spatial_rank'] == '"3dfov"':

            full_data = []
            time_concat = []
            time = len([1 for i in fdf_list if 'slab001' in i])
            for i in fdf_list: # only 1 item, but there migh be more in future
                                
                (header, data) = preproc_fdf(i)
                img_data = prepare_data(data)
                full_data.append(img_data) # full data in one list

            self.data_array = np.asarray(full_data)
            return self.data_array, self.header_dict

        #----------------------process if 2d-------------------------
        elif self.header_dict['spatial_rank'] == '"2dfov"':

            full_data = []
            time_concat = []
            time = len([1 for i in fdf_list if 'slice001' in i])

            for i in fdf_list:                
                (header, data) = preproc_fdf(i)
                img_data = prepare_data(data)
                # expand 2d to 4d
                img_data = np.expand_dims(np.expand_dims(img_data,2),3)
                full_data.append(img_data) # full data in one list
            # make sublists
            slice_list =[full_data[i:i+time] \
                            for i in range(0,len(full_data),time)]
            for aslice in slice_list:
                time_concat.append(np.concatenate(tuple(aslice),axis=3))
            #slice+time concatenated
            slice_time_concat = np.concatenate(time_concat, axis=2)

            self.data_array = slice_time_concat
            #TODO something here
            # vomit inducing method to properly flip data
            #if str(self.ppdict['orient']) == 'trans90':  
            #    self.data_array = np.swapaxes(self.data_array, 0, 1)

            return self.data_array, self.header_dict
    


        # --------------------process if 1d------------------------
        #TODO spectra
        elif self.header_dict['spatial_rank'] == '"1dfov"':
            pass
        

        return (self.header_dict, self.data_array)

    def print_header(self):
        """Print fdf header to stdout"""

        for key in self.header_dict.keys():
            print(key+' : '+str(self.header_dict[key]))
        



import vnmrjpy as vj
import numpy as np

"""
Collection of helper functions for epi and epip k-space formation
and preprocessing.
"""

def _get_navigator_echo(kspace,pd,phase_dim=1):
    """Return navigator echo from full received data"""
    if pd['navigator'] == 'y':
        navigator = kspace[:,0:1,:,:,:]
        return navigator
    else:
        return None

def _prepare_shape(kspace,pd,phase_dim=1):
    """Remove unused and navigator echo from kspace data"""
    if pd['navigator'] == 'y':
        kspace = kspace[:,2:,:,:,:]
    else:
        kspace = kspace[:,1:,:,:,:]
    return kspace

def _kzero_shift(kspace,p,phase_dim=1):
    """Shift data so first echo is at the proper space according to kzero"""

    kzero = int(p['kzero'])
    kspace = np.roll(kspace,kzero,axis=phase_dim)
    return kspace

def _reverse_even(kspace, phase_dim=1):
    """Reverse even echoes"""
    odd = kspace[:,0::2,...]
    oddflip = np.flip(odd,axis=2)
    kspace[:,0::2,...] = oddflip
    return kspace

def _reverse_odd():
    pass



import vnmrjpy as vj
import unittest
import numpy as np
import glob

class Test_getpetab(unittest.TestCase):

    def test_getpetab_fromtabfile(self):

        infile = vj.config['tablib_dir']+'/fse128_64_2'
        petab = vj.util.getpetab(infile,is_petab=True)
        self.assertEqual(len(petab.shape),2)

    def test_getpetab_fromprocpar(self):
        infile = glob.glob(vj.fids+'/fsems_s_2018111301_axial*')[0]+'/procpar'
        petab = vj.util.getpetab(infile,is_procpar=True)
        self.assertEqual(len(petab.shape),2)
        



import os

class ConfigParser():
    """Parser for vnmrjpy config file
    
    Config file is currently ./config/conf.
    Key-value pairs are simply given by key=value,
    no other structure is present yet.

    Methods:
        __init__(configfile)
        parse()
            return {key : val} dictionary

    """
    def __init__(self,configfile=None):

        default_config_path = os.path.dirname(os.path.dirname(\
                os.path.abspath(__file__)))+'/config'

        if configfile == None:
            # search for config file
            for loc in [os.curdir,\
                    os.path.expanduser("~"),\
                    default_config_path]:
                try:
                    with open(os.path.join(loc,'vnmrjpy.conf')) as conf:
                        self.configfile = os.path.join(loc,'vnmrjpy.conf')
                except:
                    pass
        else:
            self.configfile = configfile
        if self.configfile == None:
            raise(Exception('ConfigParser could not find vnmrjpy.conf'))


    def parse(self):
        """Return dictionary of config parameters"""
        conf_dict = {}
        with open(self.configfile,'r') as opencfg:
            for line in opencfg:
                line = line.rsplit('\n')[0]
                #exclude comments
                if '#' in line:
                    continue
                # key-value pairs:
                if '=' in line:
                    key = line.rsplit('=')[0]
                    val = line.rsplit('=')[1]
                    if '[' and ']' in val:
                        val_list = []
                        for item in val[1:-1].rsplit(','):
                            try:
                                item = int(item)
                            except:
                                try:
                                    item = float(item)
                                except:
                                    item = str(item)
                            val_list.append(item)
                        val = val_list
                    # check if integer
                    elif val.isdigit():
                        val = int(line.rsplit('=')[1])
                    else:
                        try:
                            val = float(val)
                        except:
                            pass
                        if val in ['False','false','FALSE']:
                            val = False
                        elif val in ['True','true','TRUE']:
                            val = True
                        else:
                            pass
                                
                    conf_dict[key] = val
        return conf_dict


if __name__ == '__main__':
    confparser = ConfigParser()
    confparser.parse()





import vnmrjpy as vj
import unittest
import numpy as np
import nibabel as nib
import glob
import copy
import matplotlib.pyplot as plt

class Test_Aloha_gems(unittest.TestCase):

    def test_gems(self):
   
        slc = 'all'  # slice
        slcnum = 'all'  #number of slices
        alohatest = vj.util.AlohaTest(slc,slcnum) 
        kspace_orig, kspace_cs, affine, procpar, savedir = \
                                        alohatest.load_test_cs_data('gems')

        #plt.imshow(np.absolute(kspace_cs[0,:,40,:,0]))
        #plt.show()        
        aloha = vj.aloha.Aloha(kspace_cs,procpar)
        kspace_filled = aloha.recon() 
        alohatest.save_test_cs_results(procpar,affine,savedir,\
                                kspace_orig,kspace_cs,kspace_filled)



import vnmrjpy as vj
import numpy as np

def calc_array_length(fid_data_shape,procpar):
    """Find the length of acquisition parameter array"""

    p = vj.io.ProcparReader(procpar).read()
    rcvrs = p['rcvrs'].count('y')
    blocks = fid_data_shape[0]
    seqcon = p['seqcon']
    # seqcon chars: echo,slice,pe1,pe2,pe3
    s_pos = [pos for pos, i in enumerate(seqcon) if i=='s']
    s_ind = np.array([0,0,0,0,0])
    # find dimensions according to seqcon
    pe = int(p['nv'])
    try:
        ne = int(p['ne'])
    except:
        ne = 1
    try:
        slc = int(p['ns'])
    except:
        slc = 1
    try:
        pe2 = int(p['nv2'])
    except:
        pe2 = 1
    # pe3 is only supported in csi
    pe3 = 1
    
    seqcon_dims = [ne,slc,pe,pe2,pe3]
    s_ind[s_pos] = 1
    blocks_wo_array = 0
    for i in range(5):
        blocks_wo_array+=seqcon_dims[i]*s_ind[i]
    if blocks_wo_array != 0:
        l = blocks // blocks_wo_array - rcvrs
    else:
        l = 1
    if l > 0:
        return l
    else:
        return 1



import vnmrjpy as vj
import nibabel as nib
import numpy as np

def make_6d(data5d, output_type='magn-phase',input_rcvrdim=5):
    """Return 6D data from complex 5D

    Input should be the usual output of KspaceMaker of ImageSpaceMaker,
    with the forat [receivers, x, y, z, time].
    Return data in [x,y,z,t,receiver,magnitude/phase] format or
    real-imaginynary

    Args:
        data5d (np.ndarray) -- input 5d data
        output_type (string) -- either 'magn-phase' or 'real-imag'
    Return:
        data6d
    """
    olddims = [0,1,2,3,4]
    newdims = [1,2,3,4,0]
    if input_rcvrdim == 0:
        data5d = np.moveaxis(data5d,newdims,olddims)
    if output_type == 'magn-phase':
        magn = np.absolute(data5d)
        phase = np.arctan2(np.imag(data5d),np.real(data5d))
        data6d = np.concatenate([magn[...,np.newaxis],phase[...,np.newaxis]],axis=-1)
    elif output_type == 'real-imag':
        real = np.real(data5d)
        imag = np.imag(data5d)
        data6d = np.concatenate([real[...,np.newaxis],imag[...,np.newaxis]],axis=-1)
    else:
        raise(Exception('Wrong "output_type" spec'))
    return data6d

def make_local_matrix(p):
    """Make matrix for nifit header based on procpar dict p

    local coords are [pahse, read, slice/phase2, time]
    """
    # the 10 multiplier is from cm -> mm conversion
    mul = 10
    if p['apptype'] in ['im2Depi']:
        dpe = float(p['lpe']) / (int(p['nphase'])) * mul
        dro = float(p['lro']) / (int(p['nread'])//2) * mul
        ds = float(p['thk'])+float(p['gap'])
        matrix = [int(p['phase']),int(p['nread'])//2,int(p['ns'])]
        dim = [dpe,dro,ds]
    elif p['apptype'] in ['im2Dfse', 'im2D', 'im2Dcs']:
        dpe = float(p['lpe']) / (int(p['nv'])) * mul
        dro = float(p['lro']) / (int(p['np'])//2) * mul
        ds = float(p['thk'])+float(p['gap'])
        matrix = [int(p['nv']),int(p['np'])//2,int(p['ns'])]
        dim = [dpe,dro,ds]
    elif p['apptype'] in ['im3D','im3Dshim', 'im3Dcs']:
        dpe = float(p['lpe']) / (int(p['nv'])) * mul
        dro = float(p['lro']) / (int(p['np'])//2) * mul
        dpe2 = float(p['lpe2']) / (int(p['nv2'])) * mul
        matrix = [int(p['nv']),int(p['np'])//2,int(p['nv2'])]
        dim = [dpe,dro,dpe2]
    
    # return dim as well?
    return matrix, dim

def swapdim(swaparr, dim):
    """Swap dimensions array to fit data created with swaparray"""

    a = sorted(zip(swaparr,dim))
    return [a[i][1] for i in range(3)]



"""
Vnmrjpy Input-Output Tools
==========================
"""
import numpy as np

from .readfid import *
from .readfdf import *
from .readprocpar import *
from .readnifti import *
from .writenifti import *
from .writefdf import *
from .saverecon import *



import unittest
import vnmrjpy as vj
import nibabel as nib
import numpy as np
import glob

class Test_anatomical(unittest.TestCase):

    def test_gems_anatomical(self):

        vj.config['verbose'] = False
        vj.config['default_space'] = 'local'
        orients = ['axial','axial90','sag','sag90','cor','cor90']
        orientdir = vj.config['dataset_dir']+'/debug/orient'
        toglob = [orientdir+'/gems*'+i+'_01*' for i in orients] 
        for num in range(len(orients)):
            seq = glob.glob(toglob[num])[0]
            print(seq)
            niftiout = vj.config['testresults_dir']+'/anatomical'
            niftiout = niftiout + '/'+orients[num]

            varr = vj.core.read_fid(seq)
            varr.to_kspace().to_imagespace().to_anatomical()
            vj.core.write.write_nifti(varr,niftiout) 


        



import unittest
import numpy as np
import vnmrjpy as vj
import time
import matplotlib.pyplot as plt
import copy

# check execution time
TEST_PERFORMANCE=True
ITER=100

def plot_test(lst):

    n = len(lst)

    for num, item in enumerate(lst):

        plt.subplot(1,n,num+1)
        plt.imshow(np.absolute(item))
        #name = [k for k,v in locals().iteritems() if v==item][0]
        plt.title(str(num))
    plt.show()

class Test_jitutils(unittest.TestCase):

    def test_avg_hankel_lvl1(self):

        pass

    def test_avg_hankel_lvl2(self):

        rp = {'fiber_shape':(4,192,192),'filter_size':(21,21),'recontype':\
                'kx-ky'}

        a = np.random.rand(*rp['fiber_shape'])
        b = np.random.rand(*rp['fiber_shape'])
        stage = 0
        fiber = np.vectorize(complex)(a,b)
        fiber = np.array(fiber,dtype='complex64')
        filter_size = rp['filter_size']
        block_shape = (fiber.shape[1]-filter_size[0]+1,filter_size[0])
        rcvrs = fiber.shape[0]
        # constructing noisy hankel
        hankel = vj.aloha.construct_lvl2_hankel(fiber,filter_size)
        noise = np.random.rand(*hankel.shape)
        hankel_noisy = hankel + noise*2
        # new averaging
        if TEST_PERFORMANCE:
            start = time.time()
            for _ in range(ITER):
                avg = vj.aloha.avg_lvl2_hankel(hankel_noisy,\
                                            block_shape,rcvrs)

            end = time.time()
            print('averaging time for hankel {}'.format((end-start)/ITER))
        else:
            avg = vj.aloha.avg_lvl2_hankel(copy.copy(hankel_noisy),\
                                            block_shape,rcvrs)
        # original averaging
        fiber_avg = vj.aloha.deconstruct_hankel(copy.copy(hankel_noisy),stage,rp)
        avg_orig = vj.aloha.construct_hankel(fiber_avg,rp)
        # difference
        diff = avg - avg_orig
        # plotting
        self.assertEqual(True,np.allclose(\
                            diff,np.zeros_like(diff), atol=1e-5))
        plot_test([hankel,hankel_noisy,avg, avg_orig, diff]) 

    def test_construct_hankel(self):
    
        rp = {'fiber_shape':(4,192,192),'filter_size':(21,21),'recontype':\
                'kx-ky'}
        stage = 0
        recontype = rp['recontype']
        recontype_int = 1  # if kx ky
        fiber_shape = rp['fiber_shape']
        a = np.random.rand(*rp['fiber_shape'])
        b = np.random.rand(*rp['fiber_shape'])
        fiber = np.vectorize(complex)(a,b)
        fiber = np.array(fiber,dtype='complex64')
        filter_size = rp['filter_size']
        hankel = vj.aloha.construct_lvl2_hankel(fiber,filter_size)
        #for _ in range(10):
        #    start = time.time()
        #    hankel = vj.aloha.construct_lvl2_hankel(fiber,filter_size)
        #    print(time.time()-start)




import unittest
import vnmrjpy as vj
import numpy as np
import matplotlib.pyplot as plt
import glob
import os


class Test_ImageSpaceMaker(unittest.TestCase):

    def test_imake_simple2d(self,plot=False,save=True):

        result_dir = vj.config['testresults_dir']+'/imake'
        if not os.path.exists(result_dir):
            os.makedirs(result_dir)
        # standard 2d
        for item in ['/gems*.fid','/sems*.fid','/mgems*.fid','/mems*.fid']:

            fiddir = sorted(glob.glob(vj.fids+item))[0]
            procpar = fiddir+'/procpar'
            fid_data, header = vj.io.FidReader(fiddir).read()
            kspace = vj.recon.KspaceMaker(\
                            fid_data, header, procpar).make()
            kspace = vj.recon.KspaceCompleter(kspace,procpar).make()
            self.assertEqual(type(kspace),np.ndarray)
            self.assertEqual(len(kspace.shape),5)
            imagespace = vj.recon.ImageSpaceMaker(kspace,procpar).make()
            if plot:
                plt.imshow(np.absolute(\
                        kspace[1,:,:,kspace.shape[3]//2,0]),\
                        cmap='gray',vmax=100)
                plt.show()
            if save:
                vj.io.SaveRecon(procpar,\
                                imagespace=imagespace)\
                                .save(result_dir,savetype='fullimag')

    def test_imake_simple3d(self,plot=False,save=True):

        result_dir = vj.config['testresults_dir']+'/imake'
        if not os.path.exists(result_dir):
            os.makedirs(result_dir)

        # standard 3d
        for item in ['/ge3d_s*.fid','/mge3d*.fid']:

            fiddir = sorted(glob.glob(vj.fids+item))[0]
            procpar = fiddir+'/procpar'
            fid_data, header = vj.io.FidReader(fiddir).read()
            kspace = vj.recon.KspaceMaker(\
                            fid_data, header, procpar).make()
            kspace = vj.recon.KspaceCompleter(kspace,procpar).make()
            self.assertEqual(type(kspace),np.ndarray)
            self.assertEqual(len(kspace.shape),5)
            imagespace = vj.recon.ImageSpaceMaker(kspace,procpar).make()

            if plot:
                plt.imshow(np.absolute(\
                        kspace[1,:,:,kspace.shape[3]//2,0]),\
                        cmap='gray',vmax=100)
                plt.show()
            if save:
                vj.io.SaveRecon(procpar,\
                                imagespace=imagespace)\
                                .save(result_dir,savetype='fullimag')

    def test_imake_ge3d_elliptical(self,plot=False,save=True):

        result_dir = vj.config['testresults_dir']+'/imake'
        if not os.path.exists(result_dir):
            os.makedirs(result_dir)

        # standard 3d
        for item in ['/ge3d_elliptical*.fid']:

            fiddir = sorted(glob.glob(vj.fids+item))[0]
            procpar = fiddir+'/procpar'
            fid_data, header = vj.io.FidReader(fiddir).read()
            kspace = vj.recon.KspaceMaker(\
                            fid_data, header, procpar).make()
            kspace = vj.recon.KspaceCompleter(kspace,procpar).make()
            self.assertEqual(type(kspace),np.ndarray)
            self.assertEqual(len(kspace.shape),5)
            imagespace = vj.recon.ImageSpaceMaker(kspace,procpar).make()

            if plot:
                plt.imshow(np.absolute(\
                        kspace[1,:,:,kspace.shape[3]//2,0]),\
                        cmap='gray',vmax=100)
                plt.show()
            if save:
                vj.io.SaveRecon(procpar,imagespace=imagespace).save(\
                                result_dir,savetype='fullimag')



import unittest
import vnmrjpy as vj
import numpy as np
import matplotlib.pyplot as plt
import time
#import cupy as cp

RP={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False}
PLOTTING = False

class Test_hankelutils(unittest.TestCase):
    
    def test_lvl2_hankel_average(self):

        #rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False,\
        #    'recontype':'k-t','fiber_shape':(4,128,21)}
        rp={'rcvrs':4,'filter_size':(21,21),'virtualcoilboost':False,\
            'recontype':'kx-ky','fiber_shape':(4,192,192)}
        stage = 0
        (p,q) = rp['filter_size']
        (rcvrs,m,n) = rp['fiber_shape']
        fiber = np.random.rand(*rp['fiber_shape'])
        
        hankel = vj.aloha.construct_hankel(fiber,rp)
        noise = np.random.rand(*hankel.shape)
        hankel_noisy = hankel+noise
        # average merthod old:
        start_old = time.time()
        fiber_avg = vj.aloha.deconstruct_hankel(hankel_noisy,stage,rp)
        hankel_avg1 = vj.aloha.construct_hankel(fiber_avg,rp)
        end_old = time.time()
        # new average method
        start_new = time.time()
        hankel_avg2 = vj.aloha.lvl2_hankel_average(hankel_noisy\
                                        ,rp['filter_size'],fiber.shape)
        end_new = time.time()
        print('old avg time: {}'.format(end_old-start_old))
        print('new avg time: {}'.format(end_new-start_new))
        plt.subplot(1,4,1)
        plt.imshow(np.real(hankel))
        plt.title('orig')
        plt.subplot(1,4,2)
        plt.imshow(np.real(hankel_avg1))
        plt.title('old avg')
        plt.subplot(1,4,3)
        plt.imshow(np.real(hankel_avg2))
        plt.title('new avg')
        plt.subplot(1,4,4)
        plt.imshow(np.real(hankel_avg2-hankel_avg1))
        plt.title('difference')
        plt.show()

    # this wont be implemented probably
    """
    def test_lvl2_hankel_weights(self):

        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False,\
            'recontype':'k-t','fiber_shape':(4,32,21)}
        filter_shape = (11,7)
        fiber_shape = (4,32,21)
        a = vj.aloha.lvl2_hankel_weights(filter_shape,fiber_shape)
    """
    # this works, but is really slow
    """
    def test_average_hankel(self):

        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False,\
            'recontype':'k-t','fiber_shape':(4,128,21),'stages':3}
        stage = 0
        hankel = np.random.rand(1770,308)
        start = time.time()
        hankel_avg = vj.aloha.average_hankel(hankel,stage,rp)
        print('hankel avg small : {}'.format(time.time()-start))

        rp={'rcvrs':4,'filter_size':(21,21),'virtualcoilboost':False,
            'recontype':'kx-ky','fiber_shape':(4,192,192),'stages':3}
        hankel = np.random.rand(29584,1764)
        start = time.time()
        hankel = vj.aloha.average_hankel(hankel,stage,rp)
        end = time.time()
        self.assertEqual(hankel.shape,(29584,1764))
        print('average big hankel time : {}'.format(end-start))
    """
    #--------------------------STANDARD TESTS----------------------------------
    def test_construct_hankel_2d(self):
        # this is the old one for time comparison
        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False}
        indata = np.random.rand(4,128,21)
        start = time.time()
        hankel = vj.aloha.construct_hankel_2d(indata,rp)
        end = time.time()
        self.assertEqual(hankel.shape,(1770,308))
        print('Construct_2d small hankel time : {}'.format(end-start))
        # test bigger kx ky
        rp={'rcvrs':4,'filter_size':(21,21),'virtualcoilboost':False}
        indata = np.random.rand(4,192,192)
        start = time.time()
        hankel = vj.aloha.construct_hankel_2d(indata,rp)
        end = time.time()
        self.assertEqual(hankel.shape,(29584,1764))
        print('Construct_2d big hankel time : {}'.format(end-start))

    def test_construct_hankel(self):

        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False}
        indata = np.random.rand(4,128,21)
        start = time.time()
        hankel = vj.aloha.construct_hankel(indata,rp)
        end = time.time()
        self.assertEqual(hankel.shape,(1770,308))
        print('Construct small hankel time : {}'.format(end-start))
        # test bigger kx ky
        rp={'rcvrs':4,'filter_size':(21,21),'virtualcoilboost':False}
        indata = np.random.rand(4,192,192)
        start = time.time()
        hankel = vj.aloha.construct_hankel(indata,rp)
        end = time.time()
        self.assertEqual(hankel.shape,(29584,1764))
        print('Construct big hankel time : {}'.format(end-start))

    def test_deconstruct_hankel(self):
        
        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False,\
            'recontype':'k-t','fiber_shape':(4,128,21)}
        hankel = np.random.rand(1770,308)
        stage = 0
        start = time.time()
        nd_data = vj.aloha.deconstruct_hankel(hankel, stage, rp)
        end = time.time()
        self.assertEqual(nd_data.shape,(4,128,21))
        print('Deconstruct small hankel time : {}'.format(end-start))

        rp={'rcvrs':4,'filter_size':(21,21),'virtualcoilboost':False,\
            'recontype':'k-t','fiber_shape':(4,192,192)}
        hankel = np.random.rand(29584,1764)
        stage = 0
        start = time.time()
        nd_data = vj.aloha.deconstruct_hankel(hankel, stage, rp)
        end = time.time()
        self.assertEqual(nd_data.shape,(4,192,192))

        print('Deconstruct big hankel time : {}'.format(end-start))

    def test_make_kspace_weights(self):
        
        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False,\
            'recontype':'k-t','fiber_shape':(4,128,21),'stages':3}
        weights = vj.aloha.make_kspace_weights(rp)
        self.assertEqual(weights[1].shape,(4,64,21))

        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':True,\
            'recontype':'kx-ky','fiber_shape':(4,128,128),'stages':3}
        weights = vj.aloha.make_kspace_weights(rp)
        self.assertEqual(weights[3].shape,(8,64,64))
        if PLOTTING == True:
            plt.imshow(np.absolute(weights[0][0,...]))
            plt.show()

    def test_init_kspace_stage(self):

        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False,\
            'recontype':'kx-ky','fiber_shape':(4,128,128),'stages':3}
        kspace = np.random.rand(4,128,128)
        stage = 1
        kspace_init = vj.aloha.init_kspace_stage(kspace,stage,rp)
        self.assertEqual(kspace_init.shape,(4,64,64))
    
        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':True,\
            'recontype':'k-t','fiber_shape':(4,128,21),'stages':3}
        kspace = np.random.rand(8,128,21)
        stage = 1
        kspace_init = vj.aloha.init_kspace_stage(kspace,stage,rp)
        self.assertEqual(kspace_init.shape,(8,64,21))

    def test_finalize_kspace_stage(self):

        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False,\
            'recontype':'kx-ky','fiber_shape':(4,128,128),'stages':3}
        fullk = np.zeros((4,128,128))
        stagek = np.ones((4,64,64))
        kspace_full = vj.aloha.finish_kspace_stage(stagek,fullk,0,rp)
        self.assertEqual(kspace_full.shape,(4,128,128))
        self.assertEqual(kspace_full[1,64,64],0)
        self.assertEqual(kspace_full[1,60,60],1)

        rp={'rcvrs':4,'filter_size':(11,7),'virtualcoilboost':False,\
            'recontype':'k-t','fiber_shape':(4,128,21),'stages':3}
        fullk = np.zeros((4,128,21))
        stagek = np.ones((4,64,21))
        kspace_full = vj.aloha.finish_kspace_stage(stagek,fullk,0,rp)
        self.assertEqual(kspace_full.shape,(4,128,21))
        self.assertEqual(kspace_full[1,60,10],1)




class ProcparReader():
    """Parses vnmrj procpar file."""

    def __init__(self, procpar_file):
        """Init procpar file reader
        
        Arguments:
        ==========
        procpar_file -- path/to/procpar
        """
        self.ppfile = str(procpar_file)

    def read(self):
        """Return a dictionary of procpar file parameters"""

        with open(self.ppfile,'r') as openpp:
            
            name = []
            value = []
            subtype = []
            basictype = []
            maxvalue = []
            minvalue = []
            # possibilities (as of manual): 0 :'line1', 1: 'line2', 2: 'last'
            line_id = 0 
        
            current_name = 0
    
            for line_num, line in enumerate(openpp.readlines()):

                if line_id == 0:  # parameter name line
                    
                    fields = line.rsplit('\n')[0].rsplit(' ')
                    current_name = fields[0]
                    name.append(fields[0])
                    subtype.append(int(fields[1]))
                    basictype.append(int(fields[2]))

                    # can be 0 (undefined), 1 (real), 2 (string)
                    current_basictype = int(fields[2])  
                    val_list = []  # if parameter value is a multitude of strings
                    line_id = 1
                                    
                elif line_id == 1:  # parameter value line

                    fields = line.rsplit('\n')[0].rsplit(' ')

                    # values are real, and all are on this line
                    if current_basictype == 1:  
                        val = [fields[i] for i in range(1,len(fields)-1)]
                        if len(val) == 1:  # don't make list if there is only one value
                            val = val[0]
                        value.append(val)
                        line_id = 2
                    # values are strings
                    elif current_basictype == 2 and int(fields[0]) == 1:
                        
                        value.append(str(fields[1].rsplit('"')[1]))
                        line_id = 2
                    # multiple string values
                    elif current_basictype == 2 and int(fields[0]) > 1:
                        
                        val_list.append(fields[1])
                        remaining_values = int(fields[0])-1
                        line_id = 3
                    
                elif line_id == 2:

                    line_id = 0
            
                elif line_id == 3:  # if values are on multiple lines
                    if remaining_values > 1:  # count backwards from remaining values
                        val_list.append(fields[0])
                        remaining_values = remaining_values - 1
                    elif remaining_values == 1:
                        val_list.append(fields[0])
                        remaining_values = remaining_values - 1
                        value.append(val_list)
                        line_id = 2
                    else:
                        print('something is off')

                else:
                    print('incorrect line_id')


        return dict(zip(name, value))
    



import vnmrjpy as vj
import numpy as np
import unittest
import glob

class Test_epip_read(unittest.TestCase):

    def test_epip_from_fid(self):

        seq = glob.glob(vj.config['fids_dir']+'/epip*1shot*')[0]
        print(seq)
        varr = vj.core.read_fid(seq)
        varr.to_kspace()



import numpy as np
import unittest
import vnmrjpy as vj
import glob
import nibabel as nib

class Test_array_param(unittest.TestCase):

    def test_calc_array_length(self):

        fid_dir = glob.glob(vj.config['dataset_dir']+'/arrayed/*arrayed_flip1*')[0]
        procpar = fid_dir+'/procpar'
        fid = fid_dir+'/fid'
        p = vj.io.ProcparReader(procpar).read()
        te_num = len(p['flip1'])
        fid_data, fid_header = vj.io.FidReader(fid,procpar).read()
        l = vj.util.calc_array_length(fid_data.shape,procpar)
        self.assertEqual(l,te_num)

    def test_array_gems_kmake(self):

        fid_dir = glob.glob(vj.config['dataset_dir']+'/arrayed/*array_flip1*')[0]
        procpar = fid_dir+'/procpar'
        fid = fid_dir+'/fid'
        image, affine = vj.io.FidReader(fid,procpar).make_image()
        ni = nib.viewers.OrthoSlicer3D(image, affine=affine)
        ni.show()
        
    
    



import vnmrjpy as vj
import numpy as np
import csv
import os

def getpetab(infile, is_procpar=None, is_petab=None):
    """Get and parse phase encode table for sequence

    Args:
        procpar (str) -- /path/to/procpar
        petab_file -- /path/to/petable_file

    Return:
        petab (np.ndarray)
    """
    def _check_result_petab(petab,params):

        pass

    def _parse_filename(name):
    
        name = name.rsplit('/')[-1]
        pass
    
    def _determine_input_type(infile):

        pass
        # just check name ...
   
    def _readpetabfile(petabfile):

        lst = []
        lst_fin = []
        if os.path.exists(petabfile):
            pass
        else:
            raise(Exception('Could not find petab file'))
        with open(petabfile, 'r') as openfile:
            reader = csv.reader(openfile,delimiter='\t')
            for item in reader:
                lst.append(item)
            # removing 't1 ='
            lst = lst[1:]
            for item in lst:
                item_fin = []
                for element in item[:-1]:  # the last one is whitespace
                    element = int(element.replace(" ",""))
                    item_fin.append(element)
                lst_fin.append(item_fin)

        petab = np.array(lst_fin)
        return petab
 
    if is_procpar == True and is_petab == None:
        filetype = 'pp'
    elif is_petab == True and is_procpar == None:
        filetype = 'petab'
    else:
        filetype = _determine_input_type(infile)
    
    if filetype == 'petab':

        # TODO check file name consistency
        return _readpetabfile(infile) 

    if filetype == 'pp':

        ppdict = vj.io.ProcparReader(infile).read()
        petabfile_name = ppdict['petable']
        try:
            fullpath = vj.config['tablib_dir']+'/'+petabfile_name
            # TODO check file name consistency
            petab = _readpetabfile(fullpath)
            return petab
        except:
            raise(Exception('could not find petab file'))



import unittest
import vnmrjpy as vj
import matplotlib.pyplot as plt
import glob

class Test_DecodeSkipint(unittest.TestCase):

    def test_decode_ge3delliptical(self):

        # check the decoding of original ge3d elliptical skipint

        #fid_dir = sorted(glob.glob(vj.fids+'/ge3d_elliptical_s_2018111301*'))[0]
        fid_dir = sorted(glob.glob(vj.fids+'/ge3d_elliptical_s_20181218*'))[1]
        procpar = fid_dir + '/procpar'
        p = vj.io.ProcparReader(procpar).read()
        skipint = p['skipint']
        decoder = vj.util.SkipintDecoder(p)
        skiptensor = decoder.make_skiptensor()
        plt.imshow(skiptensor)
        plt.show()



import copy
import numpy as np
import vnmrjpy as vj
import sys
import numba

np.set_printoptions(formatter={'float': lambda x: "{0:0.3f}".format(x)})
"""
Collection of solvers for low rank matrix completion
"""
def svt(A,known_data=None,tau=None, delta=None, epsilon=1e-4,max_iter=1000,\
        realtimeplot=False):
    """Low rank matrix completion algorihm with singular value thresholding

    Uses singular value decomposition then throws away the small singular
    values each iteration. Defaults are used as in the reference paper

    Ref.: Candes paper

    Args:
        A (np.ndarray) -- 2 dimensional numpy array to be completed
        known_data (np.ndarray)-- known elements of A in the same shape,
                                    unknown are 0
        tau (float)
        delta (float)
        epsilon (float)
        max_iter (int) -- maximum number of iterations
        realtimeplot (boolean) -- convenience option to plot the result each 
                                iteration
    Return:
        A_filled (np.ndarray) -- completed 2dim numpy array
    """
    Y = np.zeros_like(A)
    mask = np.zeros_like(A)
    mask[known_data != 0] = 1

    if tau == None:
        tau = 5*np.sum(A.shape)/2
    if delta == None:
        delta = 1.2*np.prod(A.shape)/np.sum(mask)
    if realtimeplot == True:
        rtplot = vj.util.RealTimeImshow(np.abs(A))
    
    # start iteration
   
    for _ in range(max_iter): 
        U, S, V = np.linalg.svd(Y, full_matrices=False)
        S = np.maximum(S-tau, 0)
        X = np.linalg.multi_dot([U, np.diag(S), V])
        Y = Y + delta*mask*(A-X)
    
        rel_recon_error = np.linalg.norm(mask*(X-A)) / \
                            np.linalg.norm(mask*A)

        if _ % 100 == 0:
            sys.stdout.flush()
            print(rel_recon_error)
            pass
        if rel_recon_error < epsilon:
            break
        if realtimeplot == True:
            rtplot.update_data(np.absolute(X))

    return X


def lmafit(init_data,known_data=None,\
            tol=5e-4,\
            k=1,\
            max_iter=100,\
            rank_strategy='increase',\
            verbose=False,\
            realtimeplot=False):
    """Low-rank matrix fitting algorithm

    Fills missing matrix elements by low rank approximation

    ref.: paper

    Args:
        init_data (np.ndarray) : matrix to complete. unkown elements can
                be approximated beforehand
        known_data (np.ndarray) : matrix of the same shape as init_data 
                with only the known elements, rest are zero
        tol (float) : tolerance of fitting
        rank_strategy (str) : increase or decrease rank
        verbose
        realtimeplot

    Returns:
        X (np.matrix)
        Y (np.matrix) 
        out_list [] : some list of helper outputs
    """
    (m,n) = init_data.shape
    datamask = np.zeros_like(known_data)
    datamask[known_data != 0+0*1j] = 1
    datanrm = np.linalg.norm(init_data)
    # init
    #Z = np.matrix(init_data)
    Z = init_data
    #X = np.matrix(np.zeros((m,k),dtype='complex64'))
    #TODO check removed matrix ok
    X = np.zeros((m,k),dtype='complex64')
    Y = np.eye(k,n,dtype='complex64')
    Res = np.multiply(init_data,datamask) - known_data
    res = datanrm
    reschg_tol = 0.5*tol
    # parameters for alf
    alf = 0
    increment = 0.5
    #rank estimation parameters
    itr_rank = 0
    minitr_reduce_rank = 5
    maxitr_reduce_rank = 50
    tau_limit = 10
    rank_incr = 3
    rank_max = 50

    if realtimeplot == True:
        rtplot = vj.util.RealTimeImshow(np.absolute(init_data))

    def rank_check(R,reschg,tol):
        
        # diag = np.diag(R)
        # d_hat = [diag[i]/diag[i+1] for i in range(len(diag)-1)]
        # tau = (len(diag)-1)*max(d_hat)/(sum(d_hat)-max(d_hat))

        if reschg < 10*tol:
            ind_string = 'increase'
        else:
            ind_string = 'stay'
        return ind_string

    def increase_rank(X,Y,Z,rank_incr,rank_max):
        
        k = X.shape[1]
        k_new = min(k+rank_incr,rank_max)

        m = X.shape[0]
        n = Y.shape[1]
        #X_new = np.matrix(np.zeros((m,k_new),dtype='complex64'))
        #Y_new = np.matrix(np.eye(k_new,n,dtype='complex64'))            
        X_new = np.zeros((m,k_new),dtype='complex64')
        Y_new = np.eye(k_new,n,dtype='complex64')          
        X_new[:,:k] = X
        Y_new[:k,:] = Y
        Z_new = X.dot(Y)
        return X_new, Y_new, Z_new

    # --------------MAIN ITERATION--------------------
    objv = np.zeros(max_iter)
    RR = np.ones(max_iter)

    for iter_ in range(max_iter):
        itr_rank += 1
        """
        X0 = copy.deepcopy(X)
        Y0 = copy.deepcopy(Y)
        Res0 = copy.deepcopy(Res)
        res0 = copy.deepcopy(res)
        Z0 = copy.deepcopy(Z)
        """
        X0 = X
        Y0 = Y
        Res0 = Res
        res0 = res
        Z0 = Z
        X = Z.dot(Y.conj().T)
        X, R = np.linalg.qr(X)
        Y = X.conj().T.dot(Z)
        Z = X.dot(Y)

        Res = np.multiply(known_data-Z,datamask)
        res = np.linalg.norm(Res,'fro')
        relres = res / datanrm
        ratio = res / res0
        reschg = np.abs(1-ratio)
        RR[iter_] = ratio
        # adjust alf
        if verbose == True:

            print('ratio : {}; rank : {}; reschg : {}, alf : {}'\
                    .format(ratio,X.shape[1],reschg, alf))

        if ratio >= 1.0:
            increment = np.max([0.1*alf,0.1*increment])
            X = X0
            Y = Y0
            Res =Res0
            res =res0
            relres = res / datanrm
            alf = 0
            Z = Z0
        elif ratio > 0.7:
            increment = max(increment,0.25*alf)
            alf = alf + increment 
        objv[iter_] = relres
        # check stopping
        if ((reschg < reschg_tol) and ((itr_rank > minitr_reduce_rank) \
                                or (relres < tol))):
            if verbose == True:
                print('Stopping crit achieved')
            break

        # rank adjustment
        rankadjust = rank_check(R,reschg,tol)
        if rankadjust == 'increase':
            X,Y,Z = increase_rank(X,Y,Z,rank_incr,rank_max)

        Zknown = known_data + alf*Res
        Z = Z - np.multiply(Z,datamask) + Zknown

        if realtimeplot == True:
            rtplot.update_data(np.absolute(Z))

    obj = objv[:iter_]

    return X, Y, [obj, RR, iter_, relres, reschg] 

def admm(U,V,fiber_stage_known,stage,rp,\
                                    mu=1000,\
                                    max_iter=100,\
                                    realtimeplot=False,\
                                    noiseless=True,\
                                    verbose=False):
    """Alternating Direction Method of Multipliers solver for Aloha


    Tuned for ALOHA MRI reconstruction framework, not for general use yet.
    Lmafit estimates the rank, then Admm is used to enforce the hankel structure

    refs.: Aloha papers ?
            Admm paper ?
        
    Args:
        U, V  (np.matrix) -- U*V.H is the estimated hankel matrix from Lmafit
        fiber_stage_known (np.ndarray) -- known elements in fiber,zerofilled
        stage (int) -- pyramidal decomposition stage ?? WHY??
        rp {dictionary} -- Aloha recon parameters 
        realitmeplot (boolean) -- option to plot each iteration
        noiseless (boolean) -- ?
    Returns:
        hankel
    """

    fiber_shape = fiber_stage_known.shape
    if len(fiber_shape) == 3:  # this is kx-ky, angio and k-t
        (rcvrs,m,n) = fiber_shape
        (p,q) = rp['filter_size']
        hankel_block_shape = (m-p+1,p)
        hankel_level = 2
    elif len(fiber_shape) == 2:  # this is simple k, eg.: gems
        (rcvrs,m) = fiber_shape
        p = rp['filter_size']
        hankel_block_shape = (m-p+1,p)
        hankel_level = 1
    else:
        raise(Exception('not implemented yet'))

    # init construction
    fiber_orig_part = copy.copy(fiber_stage_known)
    if hankel_level == 2:
        hankel_orig_part = vj.aloha.construct_lvl2_hankel(\
                                        fiber_orig_part,rp['filter_size'])
        hankel_mask = np.absolute(vj.aloha.construct_lvl2_hankel(\
                                        fiber_stage_known,rp['filter_size']))
    elif hankel_level == 1:
        hankel_orig_part = vj.aloha.construct_lvl1_hankel(\
                                        fiber_orig_part,rp['filter_size'])
        hankel_mask = np.absolute(vj.aloha.construct_lvl1_hankel(\
                                        fiber_stage_known,rp['filter_size']))
    #hankel_mask = np.absolute(vj.aloha.construct_hankel(fiber_stage_known,rp))
    hankel_mask[hankel_mask != 0] = 1
    hankel_mask = np.array(hankel_mask,dtype='complex64')  # where data is unknown
    hankel_mask_inv = np.ones(hankel_mask.shape) - hankel_mask

    hankel0 = U.dot(V.conj().T)
    hankel = copy.copy(hankel0)
    #hankel = U.dot(V.conj().T)

    # TODO do thios one better
    # init lagrangian update
    lagr = np.zeros(hankel.shape,dtype='complex64')
    #lagr = copy.deepcopy(hankel)
    us = (U.conj().T.dot(U)).shape
    vs = (V.conj().T.dot(V)).shape
    Iu = np.eye(us[0],us[1],dtype='complex64')
    Iv = np.eye(vs[0],vs[0],dtype='complex64')

    if verbose:
        
        res = np.linalg.norm(hankel_mask*U.dot(V.conj().T)-hankel_orig_part)
        print('mean error: {}'.format(res))

    # real time plotting for debugging purposes
    realtimeplot = realtimeplot
    if realtimeplot == True:
        rtplot = vj.util.RealTimeImshow(np.absolute(hankel0))

    for _ in range(max_iter):
    
        #start = time.time()
        # average the hankel structure, and put back original elements
        if hankel_level == 2:
            hankel = vj.aloha.avg_lvl2_hankel(hankel,hankel_block_shape,rcvrs)
        elif hankel_level == 1:
            hankel = vj.aloha.avg_lvl1_hankel(hankel,rcvrs)

        #print('admm hankel shape {}'.format(hankel.shape))
        hankel = np.multiply(hankel,hankel_mask_inv) + hankel_orig_part
        # updating U,V and the lagrangian
        #U_calc_inv = np.linalg.inv(Iv+mu*V.conj().T.dot(V))
        U = mu*(hankel0+lagr).dot(V).dot(\
                            np.linalg.inv(Iv+mu*V.conj().T.dot(V)))
        #V_calc_inv = np.linalg.inv(Iu+mu*U.conj().T.dot(U))
        V = mu*((hankel0+lagr).conj().T).dot(U).dot(\
                            np.linalg.inv(Iu+mu*U.conj().T.dot(U)))
        hankel = U.dot(V.conj().T)
        lagr = hankel0 - hankel + lagr

        if verbose:
            res = np.linalg.norm(hankel_mask*U.dot(V.conj().T)-hankel_orig_part)
            print('mean error: {}'.format(res))
        if realtimeplot == True:
            rtplot.update_data(np.absolute(hankel))
        #print('admm iter time {}'.format(time.time()-start))

    return hankel   



import os
import numpy as np
import vnmrjpy as vj

class FidReader():
    """Handles raw data from Varian spectrometer

    .fid File structure as per Vnmrj manual:
    ===================================
    struct datafilehead
    Used at the beginning of each data file (fid's, spectra, 2D) 

       int     nblocks;      /* number of blocks in file
       int     ntraces;      /* number of traces per block    
       int     np;           /* number of elements per trace    
       int     ebytes;       /* number of bytes per element            
       int     tbytes;       /* number of bytes per trace        
       int     bbytes;       /* number of bytes per block        
       short   vers_id;      /* software version and file_id status bits
       short   status;       /* status of whole file    
       int       nbheaders;     /* number of block headers            

    struct datablockhead
     Each file block contains the following header 

       short   scale;    /* scaling factor        
       short   status;    /* status of data in block      
       short   index;    /* block index              
       short   mode;    /* mode of data in block     
       int       ctcount;    /* ct value for FID        
       float   lpval;    /* F2 left phase in phasefile    
       float   rpval;    /* F2 right phase in phasefile
       float   lvl;        /* F2 level drift correction       
       float   tlt;        /* F2 tilt drift correction      
    """

    def __init__(self,fid,procpar=None):
        """Decodes .fid file header"""

        if os.path.isdir(fid):
            procpar = str(fid)+'/procpar'
            self.procpar = procpar
            fid = str(fid)+'/fid'
        else:            
            self.procpar = procpar

        def _decode_header():

            hkey_list = ['nblocks','ntraces','np','ebytes','tbytes','bbytes',\
                     'vers_id','status','nbheaders']
            hval_list = []
            h_dict = {}
            if len(self.bheader) != 32:
                print('incorrect fid header data: not 32 bytes')
                return -1
            else: 
                hval_list.append(int.from_bytes(self.bheader[0:4],byteorder='big'))
                hval_list.append(int.from_bytes(self.bheader[4:8],byteorder='big'))
                hval_list.append(int.from_bytes(self.bheader[8:12],byteorder='big'))
                hval_list.append(int.from_bytes(self.bheader[12:16],byteorder='big'))
                hval_list.append(int.from_bytes(self.bheader[16:20],byteorder='big'))
                hval_list.append(int.from_bytes(self.bheader[20:24],byteorder='big'))
                hval_list.append(int.from_bytes(self.bheader[24:26],byteorder='big'))
                hval_list.append(int.from_bytes(self.bheader[26:28],byteorder='big'))
                hval_list.append(int.from_bytes(self.bheader[28:],byteorder='big'))

                for num, i in enumerate(hkey_list):
                    h_dict[i] = hval_list[num]
            
                self.header_dict = h_dict

                def decode_status(): #TODO
                    pass

        # data should be 32 bytes
        with open(fid,'rb') as openFid:
            binary_data = openFid.read() 
            self.bheader = bytearray(binary_data[:32])
            self.bdata = binary_data[32:]

        _decode_header()

    def read(self):
        """Reads raw data, arranges into numpy array

        Return:
            data -- numpy.ndarray([blocks, data])
            header -- dictionary of fid header
    
        """

        def chunks(l, n):
            for i in range(0, len(l), n):
                yield l[i:i + n]

        chunk_size = int(self.header_dict['bbytes'])
        block_list = list(chunks(self.bdata,chunk_size))

        if self.header_dict['bbytes'] == len(block_list[0]):
            #print('blocksize check OK')
            pass

        def decode_blockhead(blockheader):
            bh_dict = {}
            bhk_list = ['scale','status','index','mode','ctcount','lpval'\
                        'rpval','lvl','tlt']
            bhv_list = []
            bhv_list.append(int.from_bytes(blockheader[0:2],byteorder='big'))
            bhv_list.append(int.from_bytes(blockheader[2:4],byteorder='big'))
            bhv_list.append(int.from_bytes(blockheader[4:6],byteorder='big'))
            bhv_list.append(int.from_bytes(blockheader[6:8],byteorder='big'))
            bhv_list.append(int.from_bytes(blockheader[8:12],byteorder='big'))
            bhv_list.append(int.from_bytes(blockheader[12:16],byteorder='big'))
            bhv_list.append(int.from_bytes(blockheader[16:20],byteorder='big'))
            bhv_list.append(int.from_bytes(blockheader[20:24],byteorder='big'))
            bhv_list.append(int.from_bytes(blockheader[24:28],byteorder='big'))

            for num, i in enumerate(bhk_list):
                bh_dict[i] = bhv_list[num]        
            self.blockhead_dict = bh_dict

        #------------main iteration through bytearray -----------------

        dim = (int(self.header_dict['nblocks']),\
                int((self.header_dict['ntraces'])*int(self.header_dict['np'])))
        DATA = np.empty(dim)
        if self.header_dict['ebytes'] == 4:
            dt = '>f'
        if self.header_dict['ebytes'] == 2:
            dt = '>i2'
        for k,block in enumerate(block_list): # for each block
            block_header = block[:28] # separate block header
            bh = decode_blockhead(block_header)
            self.print_blockheader()
            block_data = block[28:]
            DATA[k,:] = np.frombuffer(bytearray(block_data),dt)

        #self.blockhead_dict = decode_blockhead(block_header)
        return DATA, self.header_dict

    def print_header(self):
        """Prints fid header to stdout"""

        print('---------------Fid header---------------------')
        for i in sorted(self.header_dict.keys()):
            print(str(i)+' = '+str(self.header_dict[i]))

    # after read method
    def print_blockheader(self):
        
        print('---------------block header-------------------')
        for i in sorted(self.blockhead_dict.keys()):
            print(str(i)+' = '+str(self.blockhead_dict[i]))
        print('----------------------------------------------')


    
    def make_image(self):
        """Fully convert fid to 4D image as numpy array.

        This is a convenience function for testing purposes, using this
        is not advised for serious work. Relies on vj.recon mostly.

        Return:
            image (np.ndarray) -- summed image
            affine -- affine for nibabel

        """
        def _ssos(data5D):
            """Combine receivers with square root of sum of squares method"""
            ssos = np.sqrt(np.mean(np.square(np.absolute(data5D)),axis=0))            
            return ssos

        data, header = self.read()
        kspace = vj.recon.KspaceMaker(data, header, self.procpar).make()
        imgspace = vj.recon.ImageSpaceMaker(kspace, self.procpar).make()
        image = _ssos(imgspace)
        image = vj.util.to_scanner_space(image, self.procpar)
        #nwr = vj.io.NiftiWriter(image, self.procpar)
        affine = vj.util.make_scanner_affine(self.procpar)
        return image, affine
        










import numpy as np
import vnmrjpy as vj

class ImageSpaceMaker():
    """ Reconstruct MR images to real space from k-space.

    Generally this is done by fourier transform and corrections.
    Various compressed sensing approaches should be added here
    Hardcoded for each 'seqfil' sequence
    """

    def __init__(self, kspace, procpar):
        """Init params. In Aloha, standard FFT is used when k-space is ready

        Args:
            kspace = np.ndarray([receivers, phase, read, slice, echo])
            procpar = /procpar/file/path

            For CS reconstruction:
            skiptab = 'y' or 'n' procpar parameter
            skipint = procpar parameter to show which k space lines are excluded
        """
        self.kspace_data = kspace
        self.p = vj.io.ProcparReader(procpar).read()
        self.config = vj.config
        
    def make(self):
        """Standard Fourier reconstruction.

        Return imgspace = numpy.ndarray()
        """
        seqfil = str(self.p['seqfil'])
        pe_dim = self.config['pe_dim']
        pe2_dim = self.config['pe2_dim']
        ro_dim = self.config['ro_dim']

        if seqfil in ['gems', 'fsems', 'mems', 'sems', 'mgems']:

            shiftaxes = (self.config['pe_dim'],\
                        self.config['ro_dim'])
            self.kspace_data = np.fft.fftshift(self.kspace_data,axes=shiftaxes) 
            img_space = np.fft.ifft2(self.kspace_data,\
                                    axes=(pe_dim,ro_dim), norm='ortho')
            img_space = np.fft.ifftshift(img_space,\
                                    axes=(pe_dim,ro_dim))
            return img_space

        elif seqfil in ['ge3d','fsems3d','mge3d']:
            
            shiftaxes = (self.config['pe_dim'],\
                        self.config['ro_dim'],\
                        self.config['pe2_dim'])
            self.kspace_data = np.fft.fftshift(self.kspace_data,axes=shiftaxes) 
            img_space = np.fft.ifftn(self.kspace_data,\
                                    axes=(pe_dim,ro_dim,pe2_dim), norm='ortho')
            img_space = np.fft.ifftshift(img_space,\
                                    axes=(pe_dim,ro_dim,pe2_dim))
            return img_space

        elif seqfil in ['ge3d_elliptical']:

            shiftaxes = (self.config['pe_dim'],\
                        self.config['ro_dim'],\
                        self.config['pe2_dim'])
            self.kspace_data = np.fft.fftshift(self.kspace_data,axes=shiftaxes) 
            img_space = np.fft.ifftn(self.kspace_data,\
                                    axes=(pe_dim,ro_dim,pe2_dim), norm='ortho')
            img_space = np.fft.ifftshift(img_space,\
                                    axes=(pe_dim,ro_dim,pe2_dim))
            return img_space

        else:
            raise Exception('Sequence reconstruction not implemented yet')
   




import vnmrjpy as vj
import numpy as np
import glob
import nibabel as nib
import copy
"""
Collection of utility functions for systematic Aloha testing

"""

class AlohaTest():
    """This is for consistent data handling during the testphase
    """
    def __init__(self, slc, slcnum):

        # TODO
        self.slc = slc  # can be slice index, or 'center', 'offcenter'
        self.slcnum = slcnum

    def load_test_cs_fiber(self,datatype, recontype=None):
        """Load single fiber for specified recontype. 

        Helpful for realtimeplot checks.
        """
        # get testdata directory
        if datatype == 'mems':
            seqdir = '/mems_s_2018111301_axial_0_0_0_01.cs'
        elif datatype == 'angio':
            seqdir = '/ge3d_angio_HD_s_2018072604_HD_01_red8.cs'
        elif datatype == 'gems':
            seqdir = '/gems_s_2018111301_axial90_0_90_0_01.cs'
        elif datatype == 'ge3d':
            seqdir = '/ge3d_s_2018111301_axial_0_0_0_01.cs'
        elif datatpye == 'gems':
            seqdir = '/gems_s_2018111301_axial_0_0_0_01.cs'
        elif datatype == None:
            raise(Exception('Testdata not specified'))
        # get recontype
        if recontype == None:
            raise(Exception('Getting recontype here is not automatic yet'))

        # TODO recontype sanity check

         
        # same as in load_test_cs_data ... 
        testdir = vj.cs+seqdir
        procpar = testdir + '/procpar'
        resultdir = vj.config['testresults_dir']+'/aloha'+seqdir[:-3]+'.nifti'
        savedir = resultdir

        imag = []
        real = []
        imag_orig = []
        real_orig = []

        mask_img = nib.load(glob.glob(testdir+'/*mask.nii.gz')[0])
        affine = mask_img.affine
        mask = mask_img.get_fdata()

        for item in sorted(glob.glob(testdir+'/*kspace*_imag*')):
            data = nib.load(item).get_fdata()
            imag_orig.append(data)
            imag.append(np.multiply(data,mask))
        for item in sorted(glob.glob(testdir+'/*kspace*_real*')):
            data = nib.load(item).get_fdata()
            real_orig.append(data)
            real.append(np.multiply(data,mask))
        #for item in [imag,real,imag_orig,real_orig]:
        #    item = np.asarray(item)
        imag = np.asarray(imag)
        real = np.asarray(real)
        imag_orig = np.asarray(imag_orig)
        real_orig = np.asarray(real_orig)
        kspace_cs = np.vectorize(complex)(real,imag)
        kspace_orig = np.vectorize(complex)(real_orig,imag_orig)
        #
        
        # slicing k-space based on recontype
        if recontype in ['kx-ky','kx-ky_angio']:

            time = kspace_orig.shape[4]//2
            if self.slc == 'offcenter':
                ro = kspace_orig.shape[2]//2-kspace_orig.shape[2]//8
            else:
                ro = kspace_orig.shape[2]//2
            
            fiber_orig = kspace_orig[:,:,ro,:,time]
            fiber_cs = kspace_cs[:,:,ro,:,time]

        elif recontype == 'k-t':

            # take center slice
            if self.slc == 'offcenter':
                slc = kspace_orig.shape[3]//2-kspace_orig.shape[3]//8
                ro = kspace_orig.shape[2]//2-kspace_orig.shape[2]//8
            else:
                slc = kspace_orig.shape[3]//2
                ro = kspace_orig.shape[2]//2
            
            fiber_orig = kspace_orig[:,:,ro,slc,:]
            fiber_cs = kspace_cs[:,:,ro,slc,:]

        elif recontype == 'k':

            # take center slice
            if self.slc == 'offcenter':
                slc = kspace_orig.shape[3]//2-kspace_orig.shape[3]//8
                ro = kspace_orig.shape[2]//2-kspace_orig.shape[2]//8
            else:
                slc = kspace_orig.shape[3]//2
                ro = kspace_orig.shape[2]//2
            
            fiber_orig = kspace_orig[:,:,ro,slc,:]
            fiber_cs = kspace_cs[:,:,ro,slc,:]

        else:
            raise(Exception('this recontype not implemented'))

        return fiber_orig, fiber_cs


    def load_test_cs_data(self,datatype):
        """Load full data (orig, cs, affine, savepath, etc...)"""        

        SLC = self.slc
        SLCNUM = self.slcnum

        if datatype == 'mems':
            seqdir = '/mems_s_2018111301_axial_0_0_0_01.cs'
        elif datatype == 'angio':
            seqdir = '/ge3d_angio_HD_s_2018072604_HD_01_red8.cs'
        elif datatype == 'gems':
            seqdir = '/gems_s_2018111301_axial90_0_90_0_01.cs'
        elif datatype == 'ge3d':
            seqdir = '/ge3d_s_2018111301_axial_0_0_0_01.cs'
        elif datatpye == 'gems':
            seqdir = '/gems_s_2018111301_axial_0_0_0_01.cs'
        elif datatype == None:
            raise(Exception('Testdata not specified'))

        testdir = vj.cs+seqdir
        procpar = testdir + '/procpar'
        resultdir = vj.config['testresults_dir']+'/aloha'+seqdir[:-3]+'.nifti'
        savedir = resultdir

        imag = []
        real = []
        imag_orig = []
        real_orig = []

        mask_img = nib.load(glob.glob(testdir+'/*mask.nii.gz')[0])
        affine = mask_img.affine
        mask = mask_img.get_fdata()

        for item in sorted(glob.glob(testdir+'/*kspace*_imag*')):
            data = nib.load(item).get_fdata()
            imag_orig.append(data)
            imag.append(np.multiply(data,mask))
        for item in sorted(glob.glob(testdir+'/*kspace*_real*')):
            data = nib.load(item).get_fdata()
            real_orig.append(data)
            real.append(np.multiply(data,mask))
        #for item in [imag,real,imag_orig,real_orig]:
        #    item = np.asarray(item)
        imag = np.asarray(imag)
        real = np.asarray(real)
        imag_orig = np.asarray(imag_orig)
        real_orig = np.asarray(real_orig)
        kspace_cs = np.vectorize(complex)(real,imag)
        kspace_orig = np.vectorize(complex)(real_orig,imag_orig)

        if SLC == 'all':
            return (kspace_orig,\
                    kspace_cs,\
                    affine,\
                    procpar,\
                    savedir)
        
        # these are different depending on data
        elif SLC == 'center':
            if datatype in ['mems']:
                center = kspace_orig.shape[vj.config['slc_dim']]//2
                if SLCNUM > 1:
                    num1 = SLCNUM//2
                    num2 = SLCNUM//2
                else:
                    num1 = 1
                    num2 = 0
                return (kspace_orig[...,center-num1:center+num2,:],\
                        kspace_cs[...,center-num1:center+num2,:],\
                        affine,\
                        procpar,\
                        savedir)
            else:
                raise(Exception('not implemented'))
        # if slice and slice number is specified
        else:
            if datatype in ['mems']:
                return (kspace_orig[...,SLC:SLC+SLCNUM,:],\
                        kspace_cs[...,SLC:SLC+SLCNUM,:],\
                        affine,\
                        procpar,\
                        savedir)
            elif datatype in ['angio','ge3d']:
                return (kspace_orig[:,:,SLC:SLC+SLCNUM,:,:],\
                        kspace_cs[:,:,SLC:SLC+SLCNUM,:,:],\
                        affine,\
                        procpar,\
                        savedir)
            elif datatype in ['gems']:
                return (kspace_orig[...,SLC:SLC+SLCNUM,:],\
                        kspace_cs[...,SLC:SLC+SLCNUM,:],\
                        affine,\
                        procpar,\
                        savedir)
            else:
                raise(Exception('not implemented'))

    def save_test_cs_results(self,procpar,\
                                affine,\
                                savedir,\
                                kspace_orig,\
                                kspace_cs,\
                                kspace_filled):
        """Save results after ALOHA in a consistent manner"""

        k_name = ['kspace_orig','kspace_cs','kspace_filled']
        img_name = ['img_orig','img_cs','img_filled']
        dirs = ['orig','zerofilled','filled']
        for num, item in enumerate([kspace_orig,kspace_cs,kspace_filled]):
            recon= vj.recon.ImageSpaceMaker(item, procpar)
            img = recon.make()
            saverecon = vj.io.SaveRecon(procpar,kspace=item,imagespace=img)
            outdir = savedir+'/'+dirs[num]
            saverecon.save(outdir,savetype='full',filetype='nifti')




import unittest
import numpy as np
import vnmrjpy as vj
import glob
import os

class Test_niftiwriter(unittest.TestCase):

    def test_niftiwriter_ge3d(self):

        # easy 2d io
        ge3d_list = glob.glob(vj.config['dataset_dir']+'/debug/orient/ge3d*')
        ge3d = [i for i in ge3d_list if 'cor90' in i][0]
        procpar = ge3d+'/procpar'
        rdr = vj.io.FidReader(ge3d)
        image, tempaffine = rdr.make_image()
        outdir = vj.config['testresults_dir']+'/niftiwriter'
        outname = 'ge3d'
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        fullout = outdir+'/'+outname
        writer = vj.io.NiftiWriter(image,procpar,\
                    input_space='scanner',output_space='rat_anatomical')
        writer.write(fullout)

    def test_niftiwriter_gems(self):

        # easy 2d io
        gems_list = glob.glob(vj.config['dataset_dir']+'/debug/orient/gems*')
        gems = [i for i in gems_list if 'axial90' in i][0]
        procpar = gems+'/procpar'
        rdr = vj.io.FidReader(gems)
        image, tempaffine = rdr.make_image()
        outdir = vj.config['testresults_dir']+'/niftiwriter'
        outname = 'gems'
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        fullout = outdir+'/'+outname
        writer = vj.io.NiftiWriter(image,procpar,\
                    input_space='scanner',output_space='rat_anatomical')
        writer.write(fullout)

    def test_niftiwriter_gems_usual(self):

        # easy 2d io
        gems_list = glob.glob(vj.config['dataset_dir']+'/debug/orient/gems*')
        gems = [i for i in gems_list if 'usual' in i][0]
        procpar = gems+'/procpar'
        rdr = vj.io.FidReader(gems)
        image, tempaffine = rdr.make_image()
        outdir = vj.config['testresults_dir']+'/niftiwriter'
        outname = 'gems_usual'
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        fullout = outdir+'/'+outname
        writer = vj.io.NiftiWriter(image,procpar,\
                    input_space='scanner',output_space='rat_anatomical')
        writer.write(fullout)


    def test_7dnifti(self):

        mems_dir = glob.glob(vj.config['fids_dir']+'/mems*')[0]
        procpar = mems_dir+'/procpar'
        fid = mems_dir+'/fid'
        fid_data, fid_header = vj.io.FidReader(fid,procpar).read()
        kspace = vj.recon.KspaceMaker(fid_data,fid_header,procpar).make()
        imgspace = vj.recon.ImageSpaceMaker(kspace, procpar).make()
        img6d = vj.util.make_6d(imgspace,output_type='real-imag')
        writer = vj.io.NiftiWriter(img6d,procpar,input_space='local',\
                                    output_space='local')
        out6d = vj.config['testresults_dir']+'/niftiwriter/nifti6d'
        writer.write(out6d)



import vnmrjpy as vj
import unittest

class Test_config(unittest.TestCase):

    def test_configparser(self):

        conf = vj.config
        tol = conf['lmafit_tol']
        self.assertEqual(type(tol),list)
        self.assertEqual(type(vj.config['module_dir']),str)
        self.assertEqual(vj.config['rcvr_dim'],0)
        self.assertEqual(vj.config['vcboost'],bool)
